from __future__ import annotations

"""

Implements the MCTS + Self-Refine algorithm from
`Accessing GPT-4 level Mathematical Olympiad Solutions via Monte
Carlo Tree Self-refine with LLaMa-3 8B: A Technical Report`
by Zhang et. al.

The authors' [repo](https://github.com/trotsky1997/MathBlackBox) uses critiques,
refinements, and parent nodes' answers as conversation history.
I haven't tried it yet.

"""

"""

### MCTS任务构建

MCT 自优化算法代表了蒙特卡洛树搜索（MCTS）与大型语言模型的结合，将不同的场景执行任务过程抽象为搜索树结构。树上的节点代表当前不同视角下的选择策略，而边表示主体对自身的反思。该算法的操作流程遵循 MCTS 算法的一般模式。

具体来说，我们采用模型的先验知识，来让主体通过一系列的自身反思探索来获取自身对问题的最优答案；这种方式依赖模型的对齐偏好，我们在每种偏好上设计了一个工程框架，来完成自我对不同答案的奖励进行采样策略


1、对问题生成的子任务，生成一个合理的规划的节点
2、对每个节点创建一个MCTS任务，
3、输入 problem（总问题的子任务相关的子问题）
4、评分代码重构，将片段摘录器模块集成到一个关于_evaluate_answer逻辑提示模板，模板主要作用：将每个子问题相关的loader_cosplay_builder构建一个关于evaluate_system_prompt 的生成策略，具体的为编写一个关于带有评估的评估器，由loader_cosplay_builder方法返回场景执行器（CodeGeneratorBuilder），使用add_generator添加一个问答策略(CodeGenerator)中构成问答交互，build_executor后执行  executor.chat_run() 返回_ai_message

5、自我反思代码重构,将片段摘录器模块集成到一个关于self_refine逻辑提示模板，模板主要作用：将每个子问题相关的loader_cosplay_builder构建一个关于critic_system_prompt和refine_system_prompt的生成策略，critic_system_prompt为生成一个关于子问题相关的loader_cosplay_builder中自身不完美的评价内容，refine_system_prompt为不完美评价的思考过程和评分值。
具体的为编写一个关于带有评价的生成器和反思生成器，它们由loader_cosplay_builder方法返回场景执行器（CodeGeneratorBuilder），使用add_generator添加一个问答策略(CodeGenerator)中构成问答交互，build_executor后执行  executor.chat_run() 返回_ai_message



#### MCTS执行中的重要环节
MCTS中的约束规则如下，需要保证这些节点必须符合下面所定义的基本规则

提示约束：模型在奖励评分期间必须遵守最严格的标准。生成结果需要为JSON Response format
{
    "thought": "The thought process behind the answer.",
    "answer": "A float representing the answer to the problem."
}


高分抑制：评分节点中不存在满分反馈机制；任何超过 95 分的奖励都会按固定金额减少，以遏制过高分数。

重复采样：每次访问搜索树节点都涉及对节点奖励的重复采样，以增强自我评估的可靠性。需要注意的是，当对节点的子节点进行奖励采样时，我们也会对其父节点进行奖励采样，以增加奖励采样的样本量。

"""

import random
import math
from collections import deque
from enum import Enum 
from typing import Tuple
from pydantic import BaseModel
import tqdm
from dreamsboard.engine.memory.mctsr.prompt import (
    gpt_prompt_config,
    RefineResponse,
)
from dreamsboard.engine.task_engine_builder.core import CodeGeneratorBuilder
from dreamsboard.document_loaders.structured_storyboard_loader import LinkedListNode

from dreamsboard.engine.generate.code_generate import QueryProgramGenerator, BaseProgramGenerator
from langchain.schema import AIMessage
from dreamsboard.common.try_parse_json_object import try_parse_json_object
from dreamsboard.document_loaders.kor_loader import KorLoader
from dreamsboard.document_loaders.protocol.ner_protocol import TaskStepRefineNode
from langchain.prompts import PromptTemplate
from langchain_core.messages import ( 
    BaseMessage,
)
from dreamsboard.dreams.task_step_md.prompts import TASK_MD_TEMPLATE,TASK_STEP_MD_TITLE_TEMPLATE,TASK_STEP_MD_DESC_TEMPLATE,TASK_STEP_MD_LIST_TEMPLATE, TASK_STEP_MD_TEMPLATE

from langchain_core.language_models import LanguageModelInput
from langchain_core.runnables import Runnable
from dreamsboard.engine.storage.storage_context import StorageContext
from dreamsboard.document_loaders.structured_storyboard_loader import StructuredStoryboard
from dreamsboard.engine.storage.task_step_store.types import BaseTaskStepStore
from dreamsboard.common.callback import (call_func)
import numpy as np
import logging
import threading 
import re
import os

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)

logger.addHandler(handler)


PATTERN = re.compile(r"thought:\s*(.*?)\nanswer:\s*([\d.]+)", re.DOTALL)

PLAINTEXT_PATTERN = re.compile(r"```plaintext?([\s\S]*?)```[\s\S]*?answer: (\d+(\.\d+)?)", re.DOTALL)

_PROMPT_TEMPLATE_1 = PromptTemplate(input_variables=[
    "task_step_name",
    "task_step_level",
    "task_step_id",
    "task_step_question_answer"],
    template=TASK_STEP_MD_TITLE_TEMPLATE)
_PROMPT_TEMPLATE_1_1 = PromptTemplate(input_variables=[
    "task_step_name",
    "task_step_level",
    "task_step_id",
    "task_step_question_answer"],
    template=TASK_STEP_MD_DESC_TEMPLATE)
_PROMPT_TEMPLATE_1_2 = PromptTemplate(input_variables=[
    "task_step_name",
    "task_step_level",
    "task_step_id",
    "task_step_question_answer"],
    template=TASK_STEP_MD_LIST_TEMPLATE)
_PROMPT_TEMPLATE_1_3 = PromptTemplate(input_variables=[
    "task_step_name",
    "task_step_level",
    "task_step_id",
    "task_step_question_answer"],
    template=TASK_STEP_MD_TEMPLATE)
_PROMPT_TEMPLATE_2 = PromptTemplate(
    input_variables=[
        "start_task_context",
        "aemo_representation_context",
        "context_placeholder"
    ],
    template=TASK_MD_TEMPLATE)
ROOT_UCT_SCORE = 10_000


class MCTSNode(BaseModel):
    task_step_id: str
    answer: str
    """当前任务的会话信息""" 
    parent: MCTSNode | None = None
    children: list[MCTSNode] = []
    visits: int = 0
    Q: float = 0
    reward_samples: list[int] = []

    class Config:
        arbitrary_types_allowed = True

    def add_child(self, child_node: MCTSNode):
        self.children.append(child_node)

    def __repr__(self):
        return f"MCTSNode(answer={self.answer}, Q={self.Q:.2f}, visits={self.visits})"

    def add_reward(self, reward: int):
        self.reward_samples.append(reward)
        avg_reward = np.mean(self.reward_samples)
        min_reward = np.min(self.reward_samples)

        # Average worst-case and average outcomes
        self.Q = (min_reward + avg_reward) / 2


class SelectionPolicy(Enum):
    GREEDY = 1
    IMPORTANCE_SAMPLING = 2
    PAIRWISE_IMPORTANCE_SAMPLING = 3

 


class MCTSr(BaseModel):
    
    base_path: str
    task_step_id: str
    """当前任务的id"""
    storage_context: StorageContext
    """当前任务的会话存储""" 
    llm_runable: Runnable[LanguageModelInput, BaseMessage]
    problem: str
    max_rollouts: int
    exploration_constant: float = 1.0
    max_children: int = 2
    epsilon: float = 1e-10
    reward_limit: int = 95
    excess_reward_penalty: int = 5
    selection_policy: SelectionPolicy = SelectionPolicy.GREEDY 

    root: MCTSNode | None = None

    # Logs
    critiques: list[str] = []
    refinements: list[str] = []
    rewards: list[float] = []
    selected_nodes: list[MCTSNode] = []

    class Config:
        arbitrary_types_allowed = True

    @property 
    def llm_runable(self) -> Runnable[LanguageModelInput, BaseMessage]:
        return self.llm_runable

    @llm_runable.setter
    def llm_runable(self, llm_runable:Runnable[LanguageModelInput, BaseMessage]) -> None:
        self.llm_runable = llm_runable

    def self_refine(self, node: MCTSNode) -> Tuple[MCTSNode, RefineResponse]:
        raise NotImplementedError()

    def _evaluate_answer(self, node: MCTSNode) -> int:
        raise NotImplementedError()

    def self_evaluate(self, node: MCTSNode):
        """Evaluate the quality of the answer. Sample `num_samples` times and average the results."""
        reward = self._evaluate_answer(node)

        if reward > self.reward_limit:
            reward -= self.excess_reward_penalty

        node.add_reward(reward)

    def backpropagate(self, node: MCTSNode):
        parent = node.parent
        while parent:
            if parent.children is None or len(parent.children) == 0:
                parent = parent.parent
                continue

            best_child_Q = max(child.Q for child in parent.children)
            parent.Q = (parent.Q + best_child_Q) / 2
            parent.visits += 1
            parent = parent.parent

    def uct(self, node: MCTSNode):
        if not node.parent:
            # Using an arbitrarily high UCT score for the root node.
            # helps to prioritize breadth.
            return ROOT_UCT_SCORE

        return node.Q + self.exploration_constant * math.sqrt(
            math.log(node.parent.visits + 1) / (node.visits + self.epsilon)
        )

    def is_fully_expanded(self, node: MCTSNode):
        return len(node.children) >= self.max_children or any(
            child.Q > node.Q for child in node.children
        )

    def select_node(self):
        """Select a non-fully expanded node with the highest UCT value.

        A node is fully expanded if either:
        1. It has reached the max number of children
        2. Any of its children have a Q value greater than its own
        """
        candidates: list[MCTSNode] = []
        to_consider = deque([self.root])

        while to_consider:
            current_node = to_consider.popleft()
            if not self.is_fully_expanded(current_node):
                candidates.append(current_node)
            to_consider.extend(current_node.children)

        if not candidates:
            return self.root

        if self.selection_policy == SelectionPolicy.GREEDY:
            return max(candidates, key=self.uct)
        elif self.selection_policy == SelectionPolicy.IMPORTANCE_SAMPLING:
            # Sample, weighted by UCT score
            uct_scores = [self.uct(node) for node in candidates]
            selected_pair_idx = random.choices(
                range(len(candidates)), weights=uct_scores, k=1
            )[0]
            return candidates[selected_pair_idx]
        elif self.selection_policy == SelectionPolicy.PAIRWISE_IMPORTANCE_SAMPLING:
            # Sample, weighted by the difference in UCT scores between pairs
            uct_scores = [self.uct(node) for node in candidates]
            pairs = [
                (i, j) for i in range(len(candidates)) for j in range(len(candidates))
            ]
            pair_weights = [
                max(uct_scores[i], uct_scores[j]) - min(uct_scores[i], uct_scores[j])
                for i, j in pairs
            ]
            selected_pair_idx = random.choices(
                range(len(pairs)), weights=pair_weights, k=1
            )[0]
            selected_candidate_idx = max(
                pairs[selected_pair_idx], key=lambda x: uct_scores[x]
            )
            return candidates[selected_candidate_idx]
        else:
            raise ValueError(f"Invalid selection policy: {self.selection_policy}")
 

    def initialize(self):
        """Generate a zero-shot answer.""" 
        # 构建MCTS树,
        structured_storyboard = _build_structured_storyboard(self.storage_context.task_step_store)
            
        linked_list_node = structured_storyboard.get_task_step_node(self.task_step_id)
        #  初始化当前节点self.root的上级节点
        mcts_tree = linked_list_to_tree(structured_storyboard.head)

        matching_node = find_matching_node_in_tree_iterative(mcts_tree, linked_list_node)
        self.root = matching_node

    def run(self): 
          
        for _ in tqdm.tqdm(range(self.max_rollouts)):
            node = self.select_node()
            self.self_evaluate(node)
            child, refined_answer = self.self_refine(node)
            node.add_child(child)
          
            if refined_answer.answer_score > self.reward_limit:
                refined_answer.answer_score -= self.excess_reward_penalty

            child.add_reward(refined_answer.answer_score)
            self.backpropagate(child)


        owner = f"thread {threading.get_native_id()}, run end"
        logger.info(f"owner:{owner}")
        return self.get_best_answer()

    def get_best_answer(self):
        from collections import deque

        to_visit = deque([self.root])
        best_node = self.root

        while to_visit:
            current_node = to_visit.popleft()
            if current_node.Q > best_node.Q:
                best_node = current_node
            to_visit.extend(current_node.children)

        return best_node.answer

    def print(self):
        print_tree(self.root)



class MCTSrStoryboard(MCTSr): 
    """
    
    resource_id=f"resource_{node.task_step_id}"
    """
    
    @staticmethod
    def _get_ai_message(callback, resource_id, **kwargs):
 
        code_gen_builder = CodeGeneratorBuilder.from_template(nodes=[], storage_context=kwargs.get("storage_context"))
        _base_render_data = {
            'system_prompt': kwargs.get("system_prompt"),
            'messages': [kwargs.get("user_prompt")]
        }
        code_gen_builder.add_generator(BaseProgramGenerator.from_config(cfg={
            "code_file": "base_template_system.py-tpl",
            "render_data": _base_render_data,
        }))

        executor = code_gen_builder.build_executor(
            llm_runable=kwargs.get("llm_runable"),
            messages=[]
        )
        executor.execute()
        _ai_message = executor.chat_run()

        logger.info("\033[1;32m" + f"{resource_id}: {_ai_message}" + "\033[0m")
        assert executor._ai_message is not None 

        callback(_ai_message)

    @staticmethod
    def _wrapper_steps_unit(context_linked_list_node: LinkedListNode, continue_task_step_id: str):
        # 使用 TASK_STEP_MD_TEMPLATE 格式化每个任务步骤
        formatted_task_steps = [
        ]
        past_steps = []

        past_steps.append(f'{context_linked_list_node.task_step_level} task_id: {context_linked_list_node.task_step_id}, {context_linked_list_node.task_step_name}\n')

        # 计算层级关系
        level_count = context_linked_list_node.task_step_level.count('>')
        
        if level_count == 0:
            # 一级，格式化为标题 #
            step_text = _PROMPT_TEMPLATE_1.format(
                task_step_name=f'### {context_linked_list_node.task_step_name}',
                task_step_level=context_linked_list_node.task_step_level,
                task_step_description=context_linked_list_node.task_step_description,
                task_step_id=context_linked_list_node.task_step_id,
                task_step_question_answer=context_linked_list_node.task_step_question_answer
            )
            formatted_task_steps.append(step_text.strip()+"\n\n")
        elif level_count == 1:
            # 二级，格式化为标题 ##
            step_text = _PROMPT_TEMPLATE_1_1.format(
                task_step_name=f'{context_linked_list_node.task_step_name}',
                task_step_level=context_linked_list_node.task_step_level,
                task_step_description=context_linked_list_node.task_step_description,
                task_step_id=context_linked_list_node.task_step_id,
                task_step_question_answer=context_linked_list_node.task_step_question_answer
            )
            formatted_task_steps.append(step_text.strip()+"\n\n")
            
        elif level_count >= 2:
            # 三级及以上，格式化为分类 -
            step_text = _PROMPT_TEMPLATE_1_2.format(
                task_step_name=f'- {context_linked_list_node.task_step_name}',
                task_step_level=context_linked_list_node.task_step_level,
                task_step_description=context_linked_list_node.task_step_description,
                task_step_id=context_linked_list_node.task_step_id,
                task_step_question_answer=context_linked_list_node.task_step_question_answer
            )
            formatted_task_steps.append(step_text.strip()+"\n\n")

        else:
            step_text = _PROMPT_TEMPLATE_1_3.format(
                task_step_name=f'{context_linked_list_node.task_step_name}',
                task_step_level=context_linked_list_node.task_step_level,
                task_step_description=context_linked_list_node.task_step_description,
                task_step_id=context_linked_list_node.task_step_id,
                task_step_question_answer=context_linked_list_node.task_step_question_answer
            )
        
            formatted_task_steps.append(step_text.strip())

        while context_linked_list_node is not None and context_linked_list_node.next is not None:
            if context_linked_list_node.task_step_id == continue_task_step_id:
                break
            context_linked_list_node = context_linked_list_node.next
            # 计算层级关系
            level_count = context_linked_list_node.task_step_level.count('>')

            past_steps.append(f'{context_linked_list_node.task_step_level} task_id: {context_linked_list_node.task_step_id}, {context_linked_list_node.task_step_name}\n')

        
            if level_count == 0:
                # 一级，格式化为标题 #
                step_text = _PROMPT_TEMPLATE_1.format(
                    task_step_name=f'### {context_linked_list_node.task_step_name}',
                    task_step_level=context_linked_list_node.task_step_level,
                    task_step_description=context_linked_list_node.task_step_description,
                    task_step_id=context_linked_list_node.task_step_id,
                    task_step_question_answer=context_linked_list_node.task_step_question_answer
                )
                formatted_task_steps.append(step_text.strip()+"\n\n")

            elif level_count == 1:
                # 二级，格式化为标题 ##
                step_text = _PROMPT_TEMPLATE_1_1.format(
                    task_step_name=f'{context_linked_list_node.task_step_name}',
                    task_step_level=context_linked_list_node.task_step_level,
                    task_step_description=context_linked_list_node.task_step_description,
                    task_step_id=context_linked_list_node.task_step_id,
                    task_step_question_answer=context_linked_list_node.task_step_question_answer
                )
                formatted_task_steps.append(step_text.strip()+"\n\n")

            elif level_count >= 2:
                # 三级及以上，格式化为分类 -
                step_text = _PROMPT_TEMPLATE_1_2.format(
                    task_step_name=f'- {context_linked_list_node.task_step_name}',
                    task_step_level=context_linked_list_node.task_step_level,
                    task_step_description=context_linked_list_node.task_step_description,
                    task_step_id=context_linked_list_node.task_step_id,
                    task_step_question_answer=context_linked_list_node.task_step_question_answer
                )
                formatted_task_steps.append(step_text.strip()+"\n\n")

            else:
                step_text = _PROMPT_TEMPLATE_1_3.format(
                    task_step_name=f'{context_linked_list_node.task_step_name}',
                    task_step_level=context_linked_list_node.task_step_level,
                    task_step_description=context_linked_list_node.task_step_description,
                    task_step_id=context_linked_list_node.task_step_id,
                    task_step_question_answer=context_linked_list_node.task_step_question_answer
                )
            
                formatted_task_steps.append(step_text.strip())
            

        # 将格式化的步骤列表转换为字符串
        context_placeholder = "".join(formatted_task_steps)
        past_steps_placeholder = ''.join(past_steps)
        return context_placeholder, past_steps_placeholder

    def self_refine(self, node: MCTSNode) ->  Tuple[MCTSNode, RefineResponse]:
        """
        自我反思
        """

        critic_system_prompt_template =  PromptTemplate(
            input_variables=[
                "problem", 
                "current_answer",
                "context",
                "past_steps",
                "start_task_context",
                "aemo_representation_context",
                "task_step_name",
                "task_step_description",
                "task_step_level"
            ],
            template=os.environ.get(
                "critic_system_prompt_data", gpt_prompt_config.critic_system_prompt_data 
            )
        ) 

        structured_storyboard = _build_structured_storyboard(self.storage_context.task_step_store)
             
        context_linked_list_node = structured_storyboard.head
        past_context,past_steps = self._wrapper_steps_unit(context_linked_list_node, node.task_step_id)
        task_step_node = self.storage_context.task_step_store.get_task_step(self.task_step_id)
        user_prompt = critic_system_prompt_template.format(
            problem=self.problem,
            current_answer=node.answer,
            context=past_context,
            past_steps=past_steps,
            start_task_context=task_step_node.start_task_context,
            aemo_representation_context=task_step_node.aemo_representation_context,
            task_step_name=task_step_node.task_step_name,
            task_step_description=task_step_node.task_step_description,
            task_step_level=task_step_node.task_step_level
        ) 

             
        owner = f"register_event thread {threading.get_native_id()}"
        logger.info(f"owner:{owner}")
        
        results = call_func(
            self._get_ai_message,
            resource_id=f"resource_critic_{self.task_step_id}",
            kwargs={
                    "llm_runable": self.llm_runable,
                    "system_prompt": gpt_prompt_config.critic_system_prompt,
                    "user_prompt": user_prompt,
                    "storage_context": self.storage_context,
                },
        )
  
        _ai_message = results[0]
        assert _ai_message.content is not None
        critique = _ai_message.content
        assert critique is not None
        self.critiques.append(critique)
   
        refine_system_prompt_template = PromptTemplate(
            input_variables=[
                "problem", 
                "current_answer",
                "critique",
                "context",
                "past_steps",
            ],
            template=os.environ.get(
                "refine_system_prompt_data", gpt_prompt_config.refine_system_prompt_data
            )
        )
        refine_system_prompt = refine_system_prompt_template.format(
            problem=self.problem,
            current_answer=node.answer,
            critique=critique,
            context=past_context,
            past_steps=past_steps
        )
 
        owner = f"register_event thread {threading.get_native_id()}"
        logger.info(f"owner:{owner}")
        
        results = call_func(
            self._get_ai_message,
            resource_id=f"resource_refine_{node.task_step_id}",
            kwargs={
                    "llm_runable": self.llm_runable,
                    "system_prompt": gpt_prompt_config.refine_system_prompt,
                    "user_prompt": refine_system_prompt,
                    "storage_context": self.storage_context,
                },
        )
  
        _refined_answer_response_message = results[0] 
        assert _refined_answer_response_message.content is not None
            
        json_object = {}
        try:
            task_step_refine_node_list = self._kor_task_step_refine_builder(_refined_answer_response_message)
            # 将列表answer_score平均
            answer_score_list: list[float] = []
            for task_step_refine_node in task_step_refine_node_list:
                try:
                    answer_score_list.append(float(task_step_refine_node.answer_socre))
                except ValueError:
                    answer_score_list.append(0.0)
            answer_score = sum(answer_score_list) / len(answer_score_list)
            if answer_score <= 1:
                answer_score = answer_score * 100

            json_object.update({
                "thought": task_step_refine_node_list[0].thought,
                "answer": task_step_refine_node_list[0].answer,
                "answer_score": answer_score
            })
        except Exception as e: 
            json_object = {
                "thought": "解析失败",
                "answer": node.answer,
                "answer_score": 0
            }

        logger.info("\033[1;32m" + f"解析后的 JSON 对象: {json_object}" + "\033[0m")
        refined_answer = RefineResponse.model_validate(json_object)
        self.refinements.append(refined_answer)

        # ```thought \n{refined_answer.thought} ```\n\n ```answer_score \n{refined_answer.answer_score} ```
        return MCTSNode( 
            answer=f"{refined_answer.answer}",
            task_step_id=self.task_step_id, 
            parent=node,
            children=[],
            visits=0,
            Q=0,
            reward_samples=[]
        ), refined_answer

    def _evaluate_answer(self, node: MCTSNode) -> int:
        """
        评估答案
        """ 
 
        evaluate_system_prompt_template = PromptTemplate(
            input_variables=[
                "problem", 
                "past_steps", 
                "context", 
                "answer"
            ],
            template=os.environ.get(
                "evaluate_system_prompt_data", gpt_prompt_config.evaluate_system_prompt_data
            )
        )
        structured_storyboard = _build_structured_storyboard(self.storage_context.task_step_store)
             
        context_linked_list_node = structured_storyboard.head
        past_context,past_steps = self._wrapper_steps_unit(context_linked_list_node, node.task_step_id)

        user_prompt = evaluate_system_prompt_template.format(
            problem=self.problem,
            answer=node.answer,
            context=past_context,
            past_steps=past_steps
        )
        
        for attempt in range(3):
            try:
                        
                owner = f"register_event thread {threading.get_native_id()}, _evaluate_answer"
                logger.info(f"owner:{owner}")
                
                results = call_func(
                    self._get_ai_message,
                    resource_id=f"resource_evaluate_{node.task_step_id}",
                    kwargs={
                            "llm_runable": self.llm_runable,
                            "system_prompt": gpt_prompt_config.evaluate_system_prompt,
                            "user_prompt": user_prompt,
                            "storage_context": self.storage_context,
                        },
                )
         
                _ai_message = results[0]  
                
                owner = f"event end thread {threading.get_native_id()}, _evaluate_answer"
                logger.info(f"owner:{owner}")
                assert _ai_message.content is not None
                return int(_ai_message.content)
            except ValueError: 
                user_prompt = f"{_ai_message.content}\n\nFailed to parse reward as an integer."
               
                if attempt == 2:
                    raise
    
    
    def _kor_task_step_refine_builder(self, refined_answer_response_message: AIMessage) -> list[TaskStepRefineNode]:
        """
        抽取根据批评意见优化当前回答并续写上下文内容
        """
        kor_task_step_refine_builder = KorLoader.form_kor_task_step_refine_builder(self.llm_runable)
        response = kor_task_step_refine_builder.run(refined_answer_response_message.content)
        task_step_refine_node_list = []
        if response.get('data') is not None and response.get('data').get('script') is not None:
            step_list = response.get('data').get('script')
            for step in step_list:
                task_step_refine_node = TaskStepRefineNode(
                                            thought=step.get('thought'),
                                            answer=step.get('answer'),
                                            answer_socre=step.get('answer_socre')
                                            )
                task_step_refine_node_list.append(task_step_refine_node)

        return task_step_refine_node_list

def print_tree(node: MCTSNode | None, level: int = 0):
    if node is None:
        return
    indent = " " * level * 2
    node_str = repr(node)
    for line in node_str.split("\n"):
        logger.info(indent + line)
    for child in node.children:
        print_tree(child, level + 1)

 
    
def _build_structured_storyboard(task_step_store: BaseTaskStepStore) -> StructuredStoryboard:
    task_step_all = task_step_store.task_step_all
    task_step_all_list = [val.__dict__ for val in list(task_step_all.values())]
    structured_storyboard = StructuredStoryboard(json_data=task_step_all_list) 

    return structured_storyboard



# 链表转树的函数
def linked_list_to_tree(linked_list_head):
    if linked_list_head is None:
        return None

    # 递归构建树
    def build_tree(node: LinkedListNode, parent: MCTSNode = None):
        if node is None:
            return None
        
        # 为当前链表节点创建树节点
        mcts_node = MCTSNode(
            task_step_id=node.task_step_id,
            answer=node.task_step_question_answer,
            children=[], 
            visits=0, 
            Q=0, 
            reward_samples=[],
            parent=parent
        )
        
        # 如果有下一个节点，继续递归构建子节点
        if node.next:
            child_node = build_tree(node.next, mcts_node)
            mcts_node.children.append(child_node)
        
        return mcts_node

    # 从链表头开始构建树
    root = build_tree(linked_list_head)
    return root


# 使用栈来模拟深度优先搜索
def find_matching_node_in_tree_iterative(tree_node, linked_list_node):
    # 初始化栈，开始时栈中只包含根节点
    stack = [tree_node]

    while stack:
        # 获取栈顶的节点
        current_node = stack.pop()

        # 如果当前节点与链表节点相同，返回当前节点
        if current_node.answer == linked_list_node.task_step_question_answer:
            return current_node

        # 将子节点添加到栈中，按深度优先顺序处理
        for child in current_node.children:
            stack.append(child)

    return None  # 如果没有匹配的节点，返回 None


from dreamsboard.engine.constants import DATA_KEY, TYPE_KEY
from dreamsboard.engine.entity.dreams_personality.dreams_personality import DreamsPersonalityNode
from dreamsboard.engine.schema import (
    BaseNode,
)


def analysis_to_json(doc: BaseNode) -> dict:
    return {
        DATA_KEY: doc.dict(),
        TYPE_KEY: doc.get_type(),
    }


def json_to_analysis(analysis_dict: dict) -> DreamsPersonalityNode:
    analysis_type = analysis_dict[TYPE_KEY]
    data_dict = analysis_dict[DATA_KEY]
    doc: DreamsPersonalityNode

    if analysis_type == DreamsPersonalityNode.get_type():
        doc = DreamsPersonalityNode.model_validate(data_dict)
    else:
        raise ValueError(f"Unknown doc type: {analysis_type}")

    return doc



from weaviate.classes.init import AdditionalConfig, Timeout
from weaviate.classes.query import Filter
from weaviate.client import WeaviateClient
from weaviate.client import WeaviateAsyncClient, WeaviateClient
from weaviate.classes.config import Configure, Property, DataType, VectorDistances
import weaviate

import os


def init_context_connect() -> WeaviateClient:

    headers = {
        "X-Zhipuai-Api-Key": os.environ.get("ZHIPUAI_API_KEY")
    }
    client = weaviate.connect_to_local(
        headers=headers,
        additional_config=AdditionalConfig(
            timeout=Timeout(init=30, query=12000, insert=120)  # Values in seconds
        )
    )
        
    print(client.is_ready())
    return client


def init_context_collections(client: WeaviateClient, collection_id: str) -> tuple[str, str]:
    from datetime import datetime
 

    # Collection name with timestamp
    collection_name = "AtomgitPapers"+ collection_id 
    
    collection_name_context = f"{collection_name}Context"
    # client.collections.delete(collection_name) 
    # client.collections.delete(collection_name_context) 

    if not client.collections.exists(collection_name):
        client.collections.create(
            collection_name,
            reranker_config=Configure.Reranker.transformers(),
            generative_config=Configure.Generative.zhipuai(
                # These parameters are optional
                model="glm-4-plus",   
                max_tokens=500, 
                temperature=0.7,
                top_p=0.7
            ),
            vectorizer_config=[
                
                # Set another named vector
                Configure.NamedVectors.text2vec_transformers(   
                    name="chunk_text", source_properties=["chunk_text"],  
                    vector_index_config=Configure.VectorIndex.hnsw(
                        distance_metric=VectorDistances.COSINE
                    ) 
                ), 
            ],
        
            properties=[  # Define properties
                Property(name="ref_id", data_type=DataType.TEXT),
                Property(name="paper_id", data_type=DataType.TEXT),
                Property(name="paper_title", data_type=DataType.TEXT),
                Property(name="chunk_id", data_type=DataType.NUMBER),
                Property(name="chunk_text", data_type=DataType.TEXT),
                Property(name="original_filename", data_type=DataType.TEXT)
                
            ],
        )


    if not client.collections.exists(collection_name_context):
        client.collections.create(
            collection_name_context,
            
            vectorizer_config=Configure.Vectorizer.text2vec_contextionary( vectorize_collection_name=False),
        
            properties=[  # Define properties
                Property(name="refId", data_type=DataType.TEXT),
                Property(name="paperId", data_type=DataType.TEXT),
                Property(name="chunkId", data_type=DataType.NUMBER),
                Property(name="chunkText", data_type=DataType.TEXT),
                
            ],
        )

    return collection_name, collection_name_context



def insert_into_database(client: WeaviateClient, collection_name:str, union_id:str, properties_list: List[dict]):
    """
    插入数据到向量数据库,检查唯一
    """

    collection = client.collections.get(collection_name)
    union_ids = [item.get(union_id) for item in properties_list]
    response = collection.query.fetch_objects(
        filters=Filter.by_property(union_id).contains_any(union_ids),
    )

    exist_ids = [o.properties[union_id] for o in response.objects]

    for properties in properties_list:
        if properties[union_id] not in exist_ids:
            collection.data.insert(properties=properties)


from __future__ import annotations
from langchain_core.messages import ( 
    BaseMessage,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.runnables import Runnable
from langchain_core.prompts import PromptTemplate
from dreamsboard.engine.task_engine_builder.core import TaskEngineBuilder
 
from dreamsboard.document_loaders import KorLoader
from dreamsboard.dreams.aemo_representation_chain.base import AEMORepresentationChain
from dreamsboard.engine.entity.task_step.task_step import TaskStepNode
from dreamsboard.engine.storage.task_step_store.types import BaseTaskStepStore
from dreamsboard.engine.utils import concat_dirs
from dreamsboard.engine.storage.task_step_store.types import DEFAULT_PERSIST_FNAME
from dreamsboard.dreams.task_step_to_question_chain.weaviate.prepare_load import get_query_hash
import queue
import logging

import os 
import torch
from dreamsboard.vector.faiss_kb_service import FaissCollectionService
from sentence_transformers import CrossEncoder
from dreamsboard.vector.base import CollectionService
from dreamsboard.dreams.task_step_to_question_chain.weaviate.prepare_load import get_query_hash
from dreamsboard.engine.storage.task_step_store.simple_task_step_store import SimpleTaskStepStore
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)


class StructuredTaskStepStoryboard:
    """
    对任务进行规划，生成段落之间组成一个动态上下文
    任务：

        1、对任务按照提示词要求进行扩写，将扩写任务步骤收集 （src/dreamsboard/dreamsboard/engine/entity/task_step、src/dreamsboard/tests/test_kor/test_kor3.py）

        2、收集每个任务后存储到磁盘（src/dreamsboard/dreamsboard/engine/storage/task_step_store）

        3、对每个子任务载入会话场景，然后按照扩写任务步骤构建，MCTS任务 loader_task_step_iter_builder
 
    """ 
    base_path: str
    task_step_store: BaseTaskStepStore
    start_task_context: str
    cross_encoder: CrossEncoder
    collection: CollectionService
    data_base: str
    llm_runable: Runnable[LanguageModelInput, BaseMessage]
    aemo_representation_chain: AEMORepresentationChain

    def __init__(self,
                 base_path: str,
                 llm_runable: Runnable[LanguageModelInput, BaseMessage],
                 start_task_context: str,
                 cross_encoder: CrossEncoder,
                 collection: CollectionService,
                 aemo_representation_chain: AEMORepresentationChain,
                 task_step_store: BaseTaskStepStore,
                 data_base: str,
                 ):
        """

        :param base_path: 基础路径
        :param start_task_context: 开始任务
        :param aemo_representation_chain: 情感表征链
        """
        self.base_path = base_path
        self.llm_runable = llm_runable
        self.cross_encoder = cross_encoder
        self.collection = collection
        self.start_task_context = start_task_context
        self.aemo_representation_chain = aemo_representation_chain
        self.task_step_store = task_step_store
        self.data_base = data_base

    @classmethod
    def form_builder(cls,
                     llm_runable: Runnable[LanguageModelInput, BaseMessage],
                     cross_encoder_path: str,
                     embed_model_path: str,
                     start_task_context: str,
                     data_base: str = 'search_papers',
                     kor_dreams_task_step_llm: Runnable[LanguageModelInput, BaseMessage] | None = None,
                     task_step_store: BaseTaskStepStore | None = None,
                     ) -> StructuredTaskStepStoryboard: 
        aemo_representation_chain = AEMORepresentationChain.from_aemo_representation_chain(
            llm_runable=llm_runable,
            start_task_context=start_task_context,
            kor_dreams_task_step_llm=kor_dreams_task_step_llm
        )
        
        base_path = f'./{get_query_hash(start_task_context)}/'
        if not os.path.exists(base_path):
            os.makedirs(base_path)
        if task_step_store is None:
            task_step_store = SimpleTaskStepStore.from_persist_dir(f'./{base_path}/storage')
        
        collection_id = get_query_hash(start_task_context)
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        collection = FaissCollectionService(
            kb_name=collection_id,
            embed_model=embed_model_path,
            vector_name="samples",
            device=device
        )
        cross_encoder = CrossEncoder(
            cross_encoder_path,
            automodel_args={"torch_dtype": "auto"},
            trust_remote_code=True,
        )

        return cls(base_path=base_path,
                   llm_runable=llm_runable,
                   start_task_context=start_task_context,
                   cross_encoder=cross_encoder,
                   collection=collection,
                   data_base=data_base,
                   aemo_representation_chain=aemo_representation_chain,
                   task_step_store=task_step_store)


    def loader_task_step_iter_builder(self, allow_init: bool = True) -> queue.Queue[TaskEngineBuilder]:
        """
        加载任务步骤迭代器
        :param allow_init: 是否初始化
        """
        iter_builder_queue = queue.Queue()
        if not allow_init: 
            for task_step_id, task_step in self.task_step_store.task_step_all.items():
                task_step_id = task_step.node_id
                 
                # 从存储中获取详细的任务,设置到全局上下文
                task_step_store_node = SimpleTaskStepStore.from_persist_dir(f"{self.base_path}/storage/{task_step_id}")
                task_step_node = task_step_store_node.get_task_step(task_step_id)

                task_step_store_node.add_task_step([task_step_node])
                task_step_store_path = concat_dirs(dirname=f"{self.base_path}/storage/{task_step_id}", basename=DEFAULT_PERSIST_FNAME)

                self.task_step_store.persist(persist_path=task_step_store_path)
                task_step_store_path = concat_dirs(dirname=f"{self.base_path}/storage", basename=DEFAULT_PERSIST_FNAME)
                self.task_step_store.persist(persist_path=task_step_store_path) 
                iter_builder_queue.put(TaskEngineBuilder(
                    llm_runable=self.llm_runable,
                    cross_encoder=self.cross_encoder,
                    collection=self.collection,
                    start_task_context=self.start_task_context,
                    task_step_store=self.task_step_store,
                    task_step_id=task_step_id,
                    base_path=self.base_path,
                    data_base=self.data_base,
                ))

        else:
            result = self.aemo_representation_chain.invoke_aemo_representation_context()
            task_step_iter = self.aemo_representation_chain.invoke_kor_dreams_task_step_context(
                aemo_representation_context=result.get('aemo_representation_context')
            )
            for task_step in task_step_iter:
                    
                task_step_node = TaskStepNode.from_config(cfg={
                    "start_task_context": self.start_task_context,
                    "aemo_representation_context": result.get('aemo_representation_context'),
                    "task_step_name": task_step.task_step_name,
                    "task_step_description": task_step.task_step_description,
                    "task_step_level": task_step.task_step_level
                })
                task_step_id = task_step_node.node_id
  
                self.task_step_store.add_task_step([task_step_node])
                
                 
                task_step_store_path = concat_dirs(dirname=f"{self.base_path}/storage/{task_step_id}", basename=DEFAULT_PERSIST_FNAME)

                self.task_step_store.persist(persist_path=task_step_store_path)
                task_step_store_path = concat_dirs(dirname=f"{self.base_path}/storage", basename=DEFAULT_PERSIST_FNAME)
                self.task_step_store.persist(persist_path=task_step_store_path) 
                iter_builder_queue.put(TaskEngineBuilder(
                    llm_runable=self.llm_runable,
                    cross_encoder=self.cross_encoder,
                    collection=self.collection,
                    start_task_context=self.start_task_context,
                    task_step_store=self.task_step_store,
                    task_step_id=task_step_id,
                    base_path=self.base_path,
                    data_base=self.data_base,
                ))

        return iter_builder_queue



from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union
import os

import fsspec

from dreamsboard.engine.constants import (
    DOC_STORE_KEY,
    INDEX_STORE_KEY, DREAMS_ANALYSIS_STORE_KEY,
    TASK_STEP_STORE_KEY
)

from dreamsboard.engine.storage.task_step_store.simple_task_step_store import SimpleTaskStepStore
from dreamsboard.engine.storage.task_step_store.types import BaseTaskStepStore
from dreamsboard.engine.storage.task_step_store.types import DEFAULT_PERSIST_FNAME as TASK_STEP_STORE_FNAME

from dreamsboard.engine.storage.dreams_analysis_store.simple_dreams_analysis_store import SimpleDreamsAnalysisStore
from dreamsboard.engine.storage.dreams_analysis_store.types import BaseDreamsAnalysisStore
from dreamsboard.engine.storage.dreams_analysis_store.types import DEFAULT_PERSIST_FNAME as DREAMS_ANALYSIS_STORE_FNAME
from dreamsboard.engine.storage.template_store.simple_template_store import SimpleTemplateStore
from dreamsboard.engine.storage.template_store.types import DEFAULT_PERSIST_FNAME as TEMPLATE_STORE_FNAME
from dreamsboard.engine.storage.template_store.types import BaseTemplateStore
from dreamsboard.engine.storage.index_store.simple_index_store import SimpleIndexStore
from dreamsboard.engine.storage.index_store.types import (
    DEFAULT_PERSIST_FNAME as INDEX_STORE_FNAME,
)
from dreamsboard.engine.storage.index_store.types import BaseIndexStore
from dreamsboard.engine.utils import concat_dirs

DEFAULT_PERSIST_DIR = "./storage"


@dataclass
class StorageContext:
    """Storage context.

    The storage context container is a utility container for storing nodes,
    indices, and vectors. It contains the following:
    - dreams_analysis_store: BaseTemplateStore
    - template_store: BaseTemplateStore
    - index_store: BaseIndexStore

    """
    task_step_store: BaseTaskStepStore
    dreams_analysis_store: BaseDreamsAnalysisStore
    template_store: BaseTemplateStore
    index_store: BaseIndexStore

    @classmethod
    def from_defaults(
            cls,
            task_step_store: Optional[BaseTaskStepStore] = None,
            dreams_analysis_store: Optional[BaseDreamsAnalysisStore] = None,
            template_store: Optional[BaseTemplateStore] = None,
            index_store: Optional[BaseIndexStore] = None,
            persist_dir: Optional[str] = None,
            fs: Optional[fsspec.AbstractFileSystem] = None,
    ) -> "StorageContext":
        """Create a StorageContext from defaults.

        Args:
            task_step_store (Optional[BaseTaskStepStore]): task_step_store
            dreams_analysis_store (Optional[BaseDreamsAnalysisStore]): dreams_analysis_store
            template_store (Optional[BaseTemplateStore]): document store
            index_store (Optional[BaseIndexStore]): index store
            :param dreams_analysis_store:
            :param template_store:
            :param index_store:
            :param fs:
            :param persist_dir:

        """
        if persist_dir is None:
            task_step_store = task_step_store or SimpleTaskStepStore()
            dreams_analysis_store = dreams_analysis_store or SimpleDreamsAnalysisStore()
            template_store = template_store or SimpleTemplateStore()
            index_store = index_store or SimpleIndexStore()
        else:
            task_step_store = task_step_store or SimpleTaskStepStore.from_persist_dir(
                persist_dir, fs=fs
            )
            dreams_analysis_store = dreams_analysis_store or SimpleDreamsAnalysisStore.from_persist_dir(
                persist_dir, fs=fs
            )
            template_store = template_store or SimpleTemplateStore.from_persist_dir(
                persist_dir, fs=fs
            )
            index_store = index_store or SimpleIndexStore.from_persist_dir(
                persist_dir, fs=fs
            )

        return cls(task_step_store, dreams_analysis_store, template_store, index_store)

    def persist(
            self,
            persist_dir: Union[str, os.PathLike] = DEFAULT_PERSIST_DIR,
            task_step_store_fname: str = TASK_STEP_STORE_FNAME,
            dreams_analysis_store_fname: str = DREAMS_ANALYSIS_STORE_FNAME,
            template_store_fname: str = TEMPLATE_STORE_FNAME,
            index_store_fname: str = INDEX_STORE_FNAME,
            fs: Optional[fsspec.AbstractFileSystem] = None,
    ) -> None:
        """Persist the storage context.

        Args:
            persist_dir (str): directory to persist the storage context
            task_step_store_fname (str): filename for the task_step_store
            dreams_analysis_store_fname (str): filename for the dreams_analysis_store
            template_store_fname (str): filename for the template_store
            index_store_fname (str): filename for the index_store
            fs (Optional[fsspec.AbstractFileSystem]): filesystem to use

        """
        if fs is not None:
            persist_dir = str(persist_dir)  # NOTE: doesn't support Windows here
            task_step_store_path = concat_dirs(persist_dir, task_step_store_fname)
            dreams_analysis_store_path = concat_dirs(persist_dir, dreams_analysis_store_fname)
            template_store_path = concat_dirs(persist_dir, template_store_fname)
            index_store_path = concat_dirs(persist_dir, index_store_fname)
        else:
            persist_dir = Path(persist_dir)
            task_step_store_path = str(persist_dir / task_step_store_fname)
            dreams_analysis_store_path = str(persist_dir / dreams_analysis_store_fname)
            template_store_path = str(persist_dir / template_store_fname)
            index_store_path = str(persist_dir / index_store_fname)

        self.task_step_store.persist(persist_path=task_step_store_path, fs=fs)
        self.dreams_analysis_store.persist(persist_path=dreams_analysis_store_path, fs=fs)
        self.template_store.persist(persist_path=template_store_path, fs=fs)
        self.index_store.persist(persist_path=index_store_path, fs=fs)

    def to_dict(self) -> dict:
        all_simple = (
                isinstance(self.task_step_store, SimpleTaskStepStore)
                and isinstance(self.dreams_analysis_store, SimpleDreamsAnalysisStore)
                and isinstance(self.template_store, SimpleTemplateStore)
                and isinstance(self.index_store, SimpleIndexStore)
        )
        if not all_simple:
            raise ValueError(
                "to_dict only available when using simple doc/index/vector stores"
            )

        assert isinstance(self.task_step_store, SimpleTaskStepStore)
        assert isinstance(self.dreams_analysis_store, SimpleDreamsAnalysisStore)
        assert isinstance(self.template_store, SimpleTemplateStore)
        assert isinstance(self.index_store, SimpleIndexStore)

        return {
            TASK_STEP_STORE_KEY: self.task_step_store.to_dict(),
            DREAMS_ANALYSIS_STORE_KEY: self.dreams_analysis_store.to_dict(),
            DOC_STORE_KEY: self.template_store.to_dict(),
            INDEX_STORE_KEY: self.index_store.to_dict(),
        }

    @classmethod
    def from_dict(cls, save_dict: dict) -> "StorageContext":
        """Create a StorageContext from dict."""
        task_step_store = SimpleTaskStepStore.from_dict(save_dict[TASK_STEP_STORE_KEY])
        dreams_analysis_store = SimpleDreamsAnalysisStore.from_dict(save_dict[DREAMS_ANALYSIS_STORE_KEY])
        template_store = SimpleTemplateStore.from_dict(save_dict[DOC_STORE_KEY])
        index_store = SimpleIndexStore.from_dict(save_dict[INDEX_STORE_KEY])
        return cls(
            task_step_store=task_step_store,
            dreams_analysis_store=dreams_analysis_store,
            template_store=template_store,
            index_store=index_store,
        )



from abc import abstractmethod
from dataclasses import dataclass, field
from dataclasses_json import DataClassJsonMixin
from typing import Dict, List, Optional, Sequence, Set

from dreamsboard.engine.data_structs.struct_type import IndexStructType
import uuid

from dreamsboard.engine.schema import BaseNode


@dataclass
class IndexStruct(DataClassJsonMixin):
    """A base data struct for a LlamaIndex."""

    index_id: str = field(default_factory=lambda: str(uuid.uuid4()))
    summary: Optional[str] = None

    def get_summary(self) -> str:
        """Get text summary."""
        if self.summary is None:
            raise ValueError("summary field of the index_struct not set.")
        return self.summary

    @classmethod
    @abstractmethod
    def get_type(cls) -> IndexStructType:
        """Get index struct type."""


@dataclass
class IndexDict(IndexStruct):
    """A simple dictionary of documents."""

    # TODO: slightly deprecated, should likely be a list or set now
    # mapping from vector store id to node doc_id
    nodes_dict: Dict[str, str] = field(default_factory=dict)

    def add_node(
            self,
            node: BaseNode,
            text_id: Optional[str] = None,
    ) -> str:
        """Add text to table, return current position in list."""
        # # don't worry about child indices for now, nodes are all in order
        # self.nodes_dict[int_id] = node
        vector_id = text_id if text_id is not None else node.node_id
        self.nodes_dict[vector_id] = node.node_id

        return vector_id

    def delete(self, doc_id: str) -> None:
        """Delete a Node."""
        del self.nodes_dict[doc_id]

    @classmethod
    def get_type(cls) -> IndexStructType:
        """Get type."""
        return IndexStructType.CODE_GENERATOR



from dreamsboard.document_loaders.structured_storyboard_loader import (
    StructuredStoryboard,
    LinkedListNode,
    StructuredStoryboardLoader,
)
from dreamsboard.document_loaders.csv_structured_storyboard_loader import (
    StructuredStoryboardCSVBuilder
)

from dreamsboard.document_loaders.kor_loader import (
    KorLoader
)

__all__ = [
    "StructuredStoryboard",
    "LinkedListNode",
    "StructuredStoryboardLoader",
    "StructuredStoryboardCSVBuilder",
    "KorLoader"
]


def load_csv(file_dir):
    """
    获取输入文件夹内的所有txt文件，并返回文件名全称列表
    """
    import os
    L = []
    for root, dirs, files in os.walk(file_dir):
        for file in files:
            if os.path.splitext(file)[1] == '.csv':
                filename = os.path.join(root, file)
                L.append(filename)
        return L


def batch(iterable, size):
    # range对象的step是size
    for i in range(0, len(iterable), size):
        yield iterable[i:i + size]



from typing import Dict, List, Optional, Union, Any
from typing_extensions import Self
from abc import abstractmethod
from pydantic import BaseModel, Field
from enum import Enum, auto
import uuid
import json
import logging
logger = logging.getLogger(__name__)
# NOTE: for pretty printing
TRUNCATE_LENGTH = 350
WRAP_WIDTH = 70


class BaseComponent(BaseModel):
    """Base component object to caputure class names."""

    @classmethod
    @abstractmethod
    def class_name(cls) -> str:
        """Get class name."""

    def to_dict(self, **kwargs: Any) -> Dict[str, Any]:
        data = self.dict(**kwargs)
        data["class_name"] = self.class_name()
        return data

    def to_json(self, **kwargs: Any) -> str:
        data = self.to_dict(**kwargs)
        return json.dumps(data)

    # TODO: return type here not supported by current mypy version
    @classmethod
    def from_dict(cls, data: Dict[str, Any], **kwargs: Any) -> Self:  # type: ignore
        if isinstance(kwargs, dict):
            data.update(kwargs)

        data.pop("class_name", None)
        return cls(**data)

    @classmethod
    def from_json(cls, data_str: str, **kwargs: Any) -> Self:  # type: ignore
        data = json.loads(data_str)
        return cls.from_dict(data, **kwargs)


class NodeRelationship(str, Enum):
    """Node relationships used in `BaseNode` class.

    Attributes:
        SOURCE: The node is the source document.
        # PREVIOUS: The node is the previous node in the document.
        # NEXT: The node is the next node in the document.
        # PARENT: The node is the parent node in the document.
        # CHILD: The node is a child node in the document.

    """

    SOURCE = auto()
    # PREVIOUS = auto()
    # NEXT = auto()
    # PARENT = auto()
    # CHILD = auto()


class ObjectTemplateType(str, Enum):
    BaseProgramGenerator = auto()
    QueryProgramGenerator = auto()
    AIProgramGenerator = auto()
    EngineProgramGenerator = auto()


class RelatedNodeInfo(BaseComponent):
    node_id: str
    node_type: Optional[ObjectTemplateType] = None
    metadata: Dict[str, Any] = Field(default_factory=dict)
    hash: Optional[str] = None

    @classmethod
    def class_name(cls) -> str:
        """Get class name."""
        return "RelatedNodeInfo"


RelatedNodeType = Union[RelatedNodeInfo, List[RelatedNodeInfo]]


# Node classes for indexes
class BaseNode(BaseComponent):
    """Base node Object.

    Generic abstract interface for retrievable nodes

    """

    class Config:
        populate_by_name = True

    id_: str = Field(
        default_factory=lambda: str(uuid.uuid4()), description="Unique ID of the node."
    )

    """"
    metadata fields
    - injected as part of the text shown to LLMs as context
    - injected as part of the text for generating embeddings
    - used by vector DBs for metadata filtering

    """
    metadata: Dict[str, Any] = Field(
        default_factory=dict,
        description="A flat dictionary of metadata fields",
        alias="extra_info",
    )

    relationships: Dict[NodeRelationship, RelatedNodeType] = Field(
        default_factory=dict,
        description="A mapping of relationships to other node information.",
    )

    hash: str = Field(default="", description="Hash of the node content.")

    @classmethod
    @abstractmethod
    def get_type(cls) -> str:
        """Get Object type."""

    @property
    @abstractmethod
    def template_content(self) -> str:
        """Get base_template_content."""

    @template_content.setter
    @abstractmethod
    def template_content(self, _template_content) -> None:
        """Get base_template_content."""

    @property
    def node_id(self) -> str:
        return self.id_

    @node_id.setter
    def node_id(self, value: str) -> None:
        self.id_ = value

    @property
    def source_node(self) -> Optional[RelatedNodeInfo]:
        """Source object node.

        Extracted from the relationships field.

        """
        if NodeRelationship.SOURCE not in self.relationships:
            return None

        relation = self.relationships[NodeRelationship.SOURCE]
        if isinstance(relation, list):
            raise ValueError("Source object must be a single RelatedNodeInfo object")
        return relation

    @property
    def ref_template_id(self) -> Optional[str]:
        """Deprecated: Get ref doc id."""
        source_node = self.source_node
        if source_node is None:
            return None
        return source_node.node_id

    @property
    def extra_info(self) -> Dict[str, Any]:
        """TODO: DEPRECATED: Extra info."""
        return self.metadata

    def as_related_node_info(self) -> RelatedNodeInfo:
        """Get node as RelatedNodeInfo."""
        node_type = None
        try:
            node_type = ObjectTemplateType[self.get_type()]
            logger.info(f"The corresponding enum value for '{self.get_type()}' is: {node_type}")
        except KeyError:
            logger.info(f"'{self.get_type()}' is not a valid enum value.")
        return RelatedNodeInfo(
            node_id=self.node_id, node_type=node_type, metadata=self.metadata, hash=self.hash
        )



import threading
from collections import OrderedDict
from contextlib import contextmanager
from typing import Any, List, Tuple, Union, Generator

from langchain.embeddings.base import Embeddings
from langchain_community.vectorstores.faiss import FAISS
 
import logging 
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)

class ThreadSafeObject:
    def __init__(
        self, key: Union[str, Tuple], obj: Any = None, pool: "CachePool" = None
    ):
        self._obj = obj
        self._key = key
        self._pool = pool
        self._lock = threading.RLock()
        self._loaded = threading.Event()

    def __repr__(self) -> str:
        cls = type(self).__name__
        return f"<{cls}: key: {self.key}, obj: {self._obj}>"

    @property
    def key(self):
        return self._key

    @contextmanager
    def acquire(self, owner: str = "", msg: str = "") -> Generator[None, None, FAISS]:
        owner = owner or f"thread {threading.get_native_id()}"
        try:
            self._lock.acquire()
            if self._pool is not None:
                self._pool._cache.move_to_end(self.key)
            logger.info(f"{owner} 开始操作：{self.key}。{msg}")
            yield self._obj
        finally:
            logger.info(f"{owner} 结束操作：{self.key}。{msg}")
            self._lock.release()

    def start_loading(self):
        self._loaded.clear()

    def finish_loading(self):
        self._loaded.set()

    def wait_for_loading(self):
        self._loaded.wait()

    @property
    def obj(self):
        return self._obj

    @obj.setter
    def obj(self, val: Any):
        self._obj = val


class CachePool:
    def __init__(self, cache_num: int = -1):
        self._cache_num = cache_num
        self._cache = OrderedDict()
        self.atomic = threading.RLock()

    def keys(self) -> List[str]:
        return list(self._cache.keys())

    def _check_count(self):
        if isinstance(self._cache_num, int) and self._cache_num > 0:
            while len(self._cache) > self._cache_num:
                self._cache.popitem(last=False)

    def get(self, key: str) -> ThreadSafeObject:
        if cache := self._cache.get(key):
            cache.wait_for_loading()
            return cache

    def set(self, key: str, obj: ThreadSafeObject) -> ThreadSafeObject:
        self._cache[key] = obj
        self._check_count()
        return obj

    def pop(self, key: str = None) -> ThreadSafeObject:
        if key is None:
            return self._cache.popitem(last=False)
        else:
            return self._cache.pop(key, None)

    def acquire(self, key: Union[str, Tuple], owner: str = "", msg: str = ""):
        cache = self.get(key)
        if cache is None:
            raise RuntimeError(f"请求的资源 {key} 不存在")
        elif isinstance(cache, ThreadSafeObject):
            self._cache.move_to_end(key)
            return cache.acquire(owner=owner, msg=msg)
        else:
            return cache



# from .kb_api import list_kbs, create_kb, delete_kb
# from .kb_doc_api import list_docs, upload_doc, delete_doc, update_doc, download_doc, recreate_vector_store
# from .utils import KnowledgeFile, KBServiceFactory



import os
import fsspec
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from dataclasses_json import DataClassJsonMixin
from typing import Any, Dict, List, Optional, Sequence

from dreamsboard.engine.entity.dreams_personality.dreams_personality import DreamsPersonalityNode

DEFAULT_PERSIST_FNAME = "dreams_analysis_store.json"
DEFAULT_PERSIST_DIR = "./storage"
DEFAULT_PERSIST_PATH = os.path.join(DEFAULT_PERSIST_DIR, DEFAULT_PERSIST_FNAME)


@dataclass
class RefAnalysisInfo(DataClassJsonMixin):
    """Dataclass to represent ingested analyses."""

    node_ids: List = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)


class BaseDreamsAnalysisStore(ABC):
    # ===== 心里分析后开放问题、性格信息，加载与保存抽象定义Save/load =====
    def persist(
        self,
        persist_path: str = DEFAULT_PERSIST_PATH,
        fs: Optional[fsspec.AbstractFileSystem] = None,
    ) -> None:
        """Persist the docstore to a file."""
        pass

    # ===== 存储性格相关的信息 =====
    @property
    @abstractmethod
    def analysis_all(self) -> Dict[str, DreamsPersonalityNode]:
        ...

    @abstractmethod
    def add_analysis(
        self, analyses: Sequence[DreamsPersonalityNode], allow_update: bool = True
    ) -> None:
        ...

    @abstractmethod
    def get_analysis(self, analysis_id: str, raise_error: bool = True) -> Optional[DreamsPersonalityNode]:
        ...

    @abstractmethod
    def delete_analysis(self, analysis_id: str, raise_error: bool = True) -> None:
        """Delete a analysis from the store."""
        ...

    @abstractmethod
    def analysis_exists(self, analysis_id: str) -> bool:
        ...

    # ==== 存储性格相关的信息引用节点 =====
    @abstractmethod
    def get_all_ref_analysis_info(self) -> Optional[Dict[str, RefAnalysisInfo]]:
        """Get a mapping of ref_analysis_id -> RefAnalysisInfo for all ingested nodes."""

    @abstractmethod
    def get_ref_analysis_info(self, ref_analysis_id: str) -> Optional[RefAnalysisInfo]:
        """Get the RefAnalysisInfo for a given ref_analysis_id."""

    @abstractmethod
    def delete_ref_analysis(self, ref_analysis_id: str, raise_error: bool = True) -> None:
        """Delete a ref_analysis and all it's associated nodes."""

    @abstractmethod
    def ref_analysis_exists(self, ref_analysis_id: str) -> bool:
        """Check if a ref_analysis_id has been ingested."""



"""
加载csv文件
输入
空白	内容	开始时间	结束时间	分镜
七七	今天是温柔长裙风	0:00:00,560	0:00:02,720	0
七七	宝宝,你再不来我家找我玩的话	0:00:02,720	0:00:06,340	1


输出格式为
[{
start_point_len:0:00:00,560,
end_point_len:0:00:02,720,
story_board:"0",
story_board_text:"今天是温柔长裙风",
story_board_role:"七七",
},
{
start_point_len:0:00:00,560,
end_point_len:0:00:02,720,
story_board:"1",
story_board_text:"宝宝,你再不来我家找我玩的话",
story_board_role:"七七",
}]


"""
from __future__ import annotations
from abc import ABC
import pandas as pd
from typing import List, Set


class Metadata:
    def __init__(self, name, label):
        self.name = name  # 属性名称
        self.label = label  # 属性标签


class StructuredStoryboard:
    def __init__(self, start_point_len, end_point_len, story_board, story_board_text, story_board_role):
        self.start_point_len = Metadata(start_point_len, "开始时间")
        self.end_point_len = Metadata(end_point_len, "结束时间")
        self.story_board = Metadata(story_board, "分镜")
        self.story_board_text = Metadata(story_board_text, "内容")
        self.story_board_role = Metadata(story_board_role, "角色")


class StructuredStoryboardCSVBuilder(ABC):
    def __init__(self, csv_file_path: str):
        self.csv_file_path = csv_file_path
        self.data = []

    @classmethod
    def form_builder(cls, csv_file_path: str) -> StructuredStoryboardCSVBuilder:

        return cls(csv_file_path=csv_file_path)

    def export_role(self) -> Set[str]:
        """
        输出角色
        :return:
        """
        role = set()
        for item in self.data:
            if len(item.story_board_role.name) > 0 and item.story_board_role.name is not None:
                role.add(item.story_board_role.name)
        return role

    def load(self):
        # 清空现有数据
        self.data = []

        # 打开CSV文件并读取数据
        with open(self.csv_file_path, newline='', encoding='utf-8') as file:
            csv_reader = pd.read_csv(
                file, 
                dtype=str,
                keep_default_na=False,
                sep=',',
                skipinitialspace=True,
                on_bad_lines='skip',
            )
            records = csv_reader.to_dict(orient="records")
            for data_dict in records:  
                role = data_dict.get('角色', '')
                text = data_dict.get('内容', '')
                start_point_len = data_dict.get('开始时间', '')
                end_point_len = data_dict.get('结束时间', '')
                storyboard = data_dict.get('分镜', '')
                # 创建StructuredStoryboard对象
                structured_storyboard = StructuredStoryboard(start_point_len, end_point_len, storyboard, text, role)

                # 添加到数据列表
                self.data.append(structured_storyboard)

    def build_text(self, user_id: str, columns_to_select) -> str:
        """
        根据传入的列名列表筛选数据
        """
        # 根据传入的列名列表筛选数据
        selected_data = []
        for item in self.data:
            if item.story_board_role.name == user_id:
                selected_item = {}
                for column in columns_to_select:
                    if hasattr(item, column):
                        selected_item[column] = getattr(item, column).name
                selected_data.append(selected_item)

        # 获取列名对应的标签
        metadatas = []
        for column in columns_to_select:
            if hasattr(self.data[0], column):
                metadata = getattr(self.data[0], column).label
                metadatas.append(metadata)

        # 输出格式化的文本
        formatted_text = "    ".join(metadatas)
        formatted_text += "\n"
        for item in selected_data:
            # 遍历每个键值对并输出
            formatted_text += '\t'.join([item[field_name] for field_name in columns_to_select])
            if item != selected_data[-1]:
                formatted_text += "\n"

        return formatted_text

    def build_dict(self) -> dict:
        """
        把StructuredStoryboard的列表 转换为story_board字典
        (story_board:[])
        :return:
        """

        # 创建一个字典，用于按照story_board组织内容和角色
        storyboard_dict = {}

        # 遍历StructuredStoryboard对象并将其按照story_board分组
        for storyboard in self.data:
            if storyboard.story_board.name in storyboard_dict:
                # 如果已经存在该story_board的组，追加内容和角色
                if storyboard.story_board_role.name in storyboard_dict[storyboard.story_board.name]:
                    storyboard_dict[storyboard.story_board.name][storyboard.story_board_role.name].append(
                        storyboard.story_board_text.name)
                else:
                    storyboard_dict[storyboard.story_board.name][storyboard.story_board_role.name] = []
                    storyboard_dict[storyboard.story_board.name][storyboard.story_board_role.name].append(
                        storyboard.story_board_text.name)
            else:
                # 如果还没有该story_board的组，创建一个新的组

                storyboard_dict[storyboard.story_board.name] = {}
                storyboard_dict[storyboard.story_board.name][storyboard.story_board_role.name] = []
                storyboard_dict[storyboard.story_board.name][storyboard.story_board_role.name].append(
                    storyboard.story_board_text.name)

        return storyboard_dict

    def build_msg(self) -> str:
        """
        把StructuredStoryboard的列表将story_board_text和story_board_role按照story_board顺序拼接为下面内容
        story_board_role:「相同story_board的story_board_text」
        :return:
        """

        # 组织内容和角色
        formatted_text = ""

        for storyboard in self.data:
            role = storyboard.story_board_role.name
            text = storyboard.story_board_text.name
            if len(role) > 0 and len(text) > 0:
                formatted_text += f"{role}:「{''.join(text)}」\n"
            else:
                formatted_text += f"「{''.join(text)}」\n"

        return formatted_text



# -*- coding: utf-8 -*-
"""AdapterAllToolStructType class."""

from enum import Enum


class AdapterAllToolStructType(str, Enum):
    """

    Attributes:
        DICT ("dict"):

    """  

    CODE_INTERPRETER = "code_interpreter"
    DRAWING_TOOL = "drawing_tool"
    WEB_BROWSER = "web_browser"
    WEB_SEARCH = "web_search"


"""Module that contains Kor flavored encoders/decoders for CSV data.

The code will need to eventually support handling some form of nested objects,
via either JSON encoded column values or by breaking down nested attributes
into additional columns (likely both methods).
"""

from io import StringIO
from typing import Any, Dict, List

import pandas as pd

from kor.encoders.typedefs import SchemaBasedEncoder
from kor.encoders.utils import unwrap_tag, wrap_in_tag
from kor.exceptions import ParseError
from kor.nodes import AbstractSchemaNode, Object

DELIMITER = "|"


def _extract_top_level_fieldnames(node: AbstractSchemaNode) -> List[str]:
    """Temporary schema description for CSV extraction."""
    if isinstance(node, Object):
        return [attributes.id for attributes in node.attributes]
    else:
        return [node.id]


# PUBLIC API


class CSVEncoder(SchemaBasedEncoder):
    """CSV encoder."""

    def __init__(self, node: AbstractSchemaNode, use_tags: bool = False) -> None:
        """Attach node to the encoder to allow the encoder to understand schema.

        Args:
            node: The schema node to attach to the encoder.
            use_tags: Whether to wrap the output in tags. This may help identify
                      the table content in cases when the model attempts to add
                      clarifying explanations.
        """
        super().__init__(node)
        self.use_tags = use_tags

        # Verify that if we have an Object then none of its attributes are lists
        # or objects as that functionality is not yet supported.
        if isinstance(node, Object):
            for attribute in node.attributes:
                if attribute.many or isinstance(attribute, Object):
                    raise NotImplementedError(
                        "CSV Encoder does not yet support embedded lists or "
                        f"objects (attribute `{attribute.id}`)."
                    )

    def encode(self, data: Any) -> str:
        """Encode the data."""
        if not isinstance(data, dict):
            raise TypeError(f"Was expecting a dictionary got {type(data)}")

        expected_key = self.node.id

        if expected_key not in data:
            raise AssertionError(f"Expected a key: `{expected_key} to appear in data.")

        if isinstance(self.node, Object):
            field_names = _extract_top_level_fieldnames(self.node)
        else:
            field_names = [self.node.id]

        data_to_output = data[expected_key]

        if not isinstance(data_to_output, list):
            # Should always output records for pd.Dataframe
            data_to_output = [data_to_output]
        table_content = pd.DataFrame(data_to_output, columns=field_names).to_csv(
            index=False, sep=DELIMITER
        )

        if self.use_tags:
            return wrap_in_tag("csv", table_content)

        return table_content

    def decode(self, text: str) -> Dict[str, List[Dict[str, Any]]]:
        """Decode the text."""
        # First get the content between the table tags
        if self.use_tags:
            table_str = unwrap_tag("csv", text)
        else:
            table_str = text

        if table_str:
            with StringIO(table_str) as buffer:
                try:
                    df = pd.read_csv(
                        buffer,
                        dtype=str,
                        keep_default_na=False,
                        sep=DELIMITER,
                        skipinitialspace=True,
                        on_bad_lines='skip',
                    )
                except Exception as e:
                    raise ParseError(e)

            records = df.to_dict(orient="records")
        else:
            records = []

        namespace = self.node.id
        return {namespace: records}

    def get_instruction_segment(self) -> str:
        """Format instructions."""
        instructions = [
            "Please output the extracted information in CSV format in Excel dialect.",
            f"Please use a {DELIMITER} as the delimiter."
            # TODO(Eugene): Add this when we start supporting embedded columns.
            # "If a column corresponds to an array or an object,
            # use a JSON encoding to "
            # "encode its value.",
        ]

        if self.use_tags:
            instructions.append(
                "Please output a <csv> tag before and a closing </csv> after the table."
            )

        instructions.extend(
            [
                "\n",
                "Do NOT add any clarifying information.",
                "Output MUST follow the schema above.",
                "Do NOT add any additional columns that do not appear in the schema.",
            ]
        )

        return " ".join(instructions)



from llama_index.core.agent import ReActAgent
from llama_index.core.llms import LLM
from llama_index.llms.openai import OpenAI
from llama_index.core.evaluation import DatasetGenerator
from llama_index.llms.openai import OpenAI
from llama_index.core import PromptTemplate, Document


def adapter_agent(llm_agent: OpenAI, query_engine_tools):
    base_agent = ReActAgent.from_tools(query_engine_tools, llm=llm_agent, verbose=True)
    return base_agent


def build_dataset_generator_questions(base_question_gen_query, docs: list[Document], llm: LLM, num=60) -> list[str]:

    # 在
    # /llama_index/core/evaluation/dataset_generation.py:246
    # /llama_index/core/evaluation/dataset_generation.py:274
    #  增加延时函数
    #  await asyncio.sleep(0.5)
    dataset_generator = DatasetGenerator.from_documents(
        docs,
        question_gen_query=base_question_gen_query,
        llm=llm,
        show_progress=True,
    )

    questions = dataset_generator.generate_questions_from_nodes(num=num)
    return questions


def gen_question_variations(vary_question_tmpl, valid_10q_str, generator_questions, llm: LLM, num_vary=3) -> list[str]:

    prompt_tmpl = PromptTemplate(vary_question_tmpl)

    new_questions = []
    for idx, question in enumerate(generator_questions):
        new_questions.append(question)
        response = llm.complete(
            prompt_tmpl.format(
                num_vary=num_vary,
                base_question=question,
                valid_10qs=valid_10q_str,
            )
        )
        # parse into newlines
        raw_lines = str(response).split("\n")
        cur_new_questions = [l for l in raw_lines if l != ""]
        print(f"[{idx}] Original Question: {question}")
        print(f"[{idx}] Generated Question Variations: {cur_new_questions}")
        new_questions.extend(cur_new_questions)

    return new_questions



from __future__ import annotations
from abc import ABC
from langchain_core.messages import (
    BaseMessage,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.runnables import Runnable
from langchain.chains.base import Chain
from langchain_core.prompts  import PromptTemplate
from langchain.chains import SequentialChain
from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnableParallel, RunnableLambda

from dreamsboard.document_loaders import KorLoader

from dreamsboard.engine.storage.storage_context import BaseTaskStepStore

from dreamsboard.document_loaders.protocol.ner_protocol import TaskStepNode
from dreamsboard.dreams.task_step_to_question_chain.prompts import (
    CONVERT_TASK_STEP_TO_QUESTION_PROMPT_TEMPLATE,
    TASK_STEP_QUESTION_TO_GRAPHQL_PROMPT_TEMPLATE
)

from dreamsboard.engine.entity.task_step.task_step import TaskStepContext

from dreamsboard.dreams.task_step_to_question_chain.weaviate.prepare_load import exe_query, get_query_hash
from dreamsboard.dreams.task_step_to_question_chain.searx.searx import searx_query

import threading
from dreamsboard.common.callback import (event_manager)
from sentence_transformers import CrossEncoder
from dreamsboard.engine.storage.task_step_store.types import DEFAULT_PERSIST_FNAME
from dreamsboard.engine.utils import concat_dirs
from typing import List, Dict
from dreamsboard.vector.base import CollectionService, DocumentWithVSId
import logging
import os
import re
import pandas as pd
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)

"""
#### 场景加载模块

编写符合计算机科学领域的 故事情境提示词，生成研究情境（story_scenario_context），替换现有的langchain会话模板，
对每个子任务指令转换为子问题
召回问题前3条,存入task_step_question_context
调用llm，生成task_step_question_answer
"""

PATTERN = re.compile(r"```graphql?([\s\S]*?)```", re.DOTALL)
"""Regex pattern to parse the output."""

class TaskStepToQuestionChain(ABC):
    base_path: str
    start_task_context: str
    task_step_to_question_chain: Chain
    task_step_store: BaseTaskStepStore
    task_step_question_to_graphql_chain: Chain
    collection: CollectionService
    cross_encoder: CrossEncoder
    data_base: str
    def __init__(self,
                 base_path: str,
                 start_task_context: str,
                 task_step_store: BaseTaskStepStore,
                 task_step_to_question_chain: Chain,
                 task_step_question_to_graphql_chain: Chain,
                 collection: CollectionService,
                 cross_encoder: CrossEncoder,
                 data_base: str,
                 ):
        self.base_path = base_path
        self.start_task_context = start_task_context
        self.task_step_store = task_step_store
        self.task_step_to_question_chain = task_step_to_question_chain
        self.task_step_question_to_graphql_chain = task_step_question_to_graphql_chain
        self.collection = collection
        self.cross_encoder = cross_encoder
        self.data_base = data_base

    @classmethod
    def from_task_step_to_question_chain(
            cls,
            base_path: str,
            start_task_context: str,
            llm_runable: Runnable[LanguageModelInput, BaseMessage],
            task_step_store: BaseTaskStepStore,
            collection: CollectionService,
            cross_encoder: CrossEncoder,
            data_base: str = 'search_papers',
    ) -> TaskStepToQuestionChain:
        """

        1、对当前主任务名称创建hash值，作为collection_name
        2、对当前任务步骤名称创建hash值，作为collection_name_context
        """
        prompt_template1 = PromptTemplate(input_variables=["start_task_context",
                                                           "aemo_representation_context",
                                                           "task_step_name",
                                                           "task_step_description",
                                                           "task_step_level"],
                                          template=os.environ.get(
                                              "CONVERT_TASK_STEP_TO_QUESTION_PROMPT_TEMPLATE", CONVERT_TASK_STEP_TO_QUESTION_PROMPT_TEMPLATE
                                          ))

        task_step_to_question_chain = (prompt_template1 | llm_runable | StrOutputParser())

        def wrapper_output(_dict):
            return {
                # 中间变量全部打包输出
                "task_step_question_context": _dict["task_step_question_context"],
            }

        task_step_to_question_chain = ({
                                           "task_step_question_context": task_step_to_question_chain,
                                       }
                                       | RunnableLambda(wrapper_output))

        prompt_template2 = PromptTemplate(input_variables=[
            "collection_name_context",
            "task_step_question"
        ],
            template=os.environ.get("TASK_STEP_QUESTION_TO_GRAPHQL_PROMPT_TEMPLATE", TASK_STEP_QUESTION_TO_GRAPHQL_PROMPT_TEMPLATE)
        )

        task_step_question_to_graphql_chain = (prompt_template2 | llm_runable | StrOutputParser())

        def wrapper_output2(_dict):
            return {
                # 中间变量全部打包输出
                "task_step_question_graphql_context": _dict["task_step_question_graphql_context"],
            }

        task_step_question_to_graphql_chain = ({
                                                   "task_step_question_graphql_context": task_step_question_to_graphql_chain,
                                               }
                                               | RunnableLambda(wrapper_output2))

        return cls(
            base_path=base_path,
            start_task_context=start_task_context,
            task_step_store=task_step_store,
            task_step_to_question_chain=task_step_to_question_chain,
            task_step_question_to_graphql_chain=task_step_question_to_graphql_chain,
            collection=collection,
            cross_encoder=cross_encoder,
            data_base=data_base,
        )

    def invoke_task_step_to_question(self, task_step_id: str) -> None:
        """
        对开始任务进行抽取，得到任务步骤
        :return:
        """

        task_step_node = self.task_step_store.get_task_step(task_step_id)
        result = self.task_step_to_question_chain.invoke(task_step_node.__dict__)
        task_step_node.task_step_question = result["task_step_question_context"]
        self.task_step_store.add_task_step([task_step_node])

        # 每处理一个任务步骤，就持久化一次
        task_step_store_path = concat_dirs(dirname=f"{self.base_path}/storage/{task_step_id}", basename=DEFAULT_PERSIST_FNAME)
        self.task_step_store.persist(persist_path=task_step_store_path)

    @staticmethod
    def _into_database_query(callback, resource_id, **kwargs) -> None:
        """
        插入数据到向量数据库,检查唯一
        :param union_id_key:  唯一标识
        :param page_content_key:  数据列表
        :param properties_list:  数据列表 
        :return: None
        """

        union_ids = [str(item.get(kwargs.get("union_id_key"))) for item in kwargs.get("properties_list")]

        response = kwargs.get("collection").get_doc_by_ids(ids=union_ids)

        exist_ids = [o.metadata[kwargs.get("union_id_key")] for o in response]

        docs = []
        for item in kwargs.get("properties_list"):
            metadata = {key: value for key, value in item.items() if key != kwargs.get("page_content_key")}
            if item.get(kwargs.get("union_id_key")) not in exist_ids:
                doc = DocumentWithVSId(id=item.get(kwargs.get("union_id_key")), page_content=item.get(kwargs.get("page_content_key")), metadata=metadata )
                docs.append(doc)

        kwargs.get("collection").do_add_doc(docs)
        
        kwargs.get("collection").save_vector_store()
        # 召回 
        response = kwargs.get("collection").do_search(query=kwargs.get('task_step_question'), top_k=10, score_threshold=0.6)
        callback(response)


    def invoke_task_step_question_context(self, task_step_id: str) -> None:
        """
        对任务步骤进行抽取，得到任务步骤的上下文
        3、增加项目
        """

        from langchain_community.utilities import SearxSearchWrapper
        task_step_node = self.task_step_store.get_task_step(task_step_id)
        if task_step_node.task_step_question_context is not None and len(task_step_node.task_step_question_context) > 0:
            return

        top_k =30  # 可选参数，默认查询返回前 30个结果
        if self.data_base == 'search_papers':

            properties_list = exe_query(task_step_node.task_step_question,top_k)
        elif self.data_base == 'searx':

            properties_list = searx_query(task_step_node.task_step_question,top_k)

        # 插入数据到数据库
        owner = f"register_event thread {threading.get_native_id()}"
        logger.info(f"owner:{owner}")
        event_id = event_manager.register_event(
            self._into_database_query,
            resource_id=f"resource_collection_{self.collection.kb_name}",
            kwargs={
                    "collection": self.collection,
                    "union_id_key": 'ref_id',
                    "page_content_key": 'chunk_text',
                    "properties_list": properties_list,
                    "task_step_question": task_step_node.task_step_question,
                },
        ) 
        results = None
        while results is None or len(results) == 0:
            results = event_manager.get_results(event_id) 
        response = results[0]

        chunk_texts = []
        ref_ids = []
        chunk_ids = []
        for o in response:
            chunk_texts.append(o.page_content)
            ref_ids.append(o.metadata['ref_id'])
            chunk_ids.append(o.metadata['chunk_id'])

        rankings = self.cross_encoder.rank(
            task_step_node.task_step_question,
            chunk_texts,
            show_progress_bar=True,
            return_documents=True,
            convert_to_tensor=True
        )

        task_step_question_context = []
        # 召回问题前3条，存入task_step_question_context
        for i, ranking in enumerate(rankings[:3]):
            logger.info(f"ref_ids: {ref_ids[i]},chunk_ids: {chunk_ids[i]}, Score: {ranking['score']:.4f}, Text: {ranking['text']}")
            task_step_question_context.append(TaskStepContext(ref_id=str(ref_ids[i]), chunk_id=str(chunk_ids[i]), score=ranking['score'], text=ranking['text']))

        # rankingscorpus_id 的索引与ref_ids的索引一致 
        task_step_node.task_step_question_context = task_step_question_context
        self.task_step_store.add_task_step([task_step_node])

        # 每处理一个任务步骤，就持久化一次
        task_step_store_path = concat_dirs(dirname=f"{self.base_path}/storage/{task_step_id}", basename=DEFAULT_PERSIST_FNAME)
        self.task_step_store.persist(persist_path=task_step_store_path)


    def export_csv_file_path(self, task_step_id: str) -> str:
        """
        3、对召回内容与问题 导出csv文件
        """
        task_step_node = self.task_step_store.get_task_step(task_step_id)

        table_data = []
        row = [
            task_step_id,
            task_step_node.task_step_name,
            task_step_node.task_step_level
        ]

        table_data.append(row)
        row2 = [
            task_step_id,
            task_step_node.task_step_question,
            task_step_node.task_step_level
        ]
        table_data.append(row2)
        for context in task_step_node.task_step_question_context:
            row3 = [
                task_step_id,
                f"ref_ids: {context['ref_id']}, chunk_ids: {context['chunk_id']}, Score: {context['score']:.4f}, Text: {context['text']}",
                task_step_node.task_step_level
            ]
            table_data.append(row3)

        table = pd.DataFrame(table_data, columns=["角色", "内容", "分镜"])

        table.to_csv(f"{self.base_path}/storage/{task_step_id}.csv", index=False, escapechar='\\')

        return f"{self.base_path}/storage/{task_step_id}.csv"


from langchain_community.utilities import SearxSearchWrapper

from typing import List

import requests

import hashlib
import os
import json
import tempfile 
import logging  

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)

CACHE_DIR = os.path.join(tempfile.gettempdir(), "query_cache")  # 自定义缓存目录
 
def ensure_cache_dir():
    """确保缓存目录存在"""
    if not os.path.exists(CACHE_DIR):
        os.makedirs(CACHE_DIR)

def get_query_hash(query):
    """根据 query 生成唯一的哈希值"""
    return hashlib.md5(query.encode('utf-8')).hexdigest()

def get_cache_file_path(query):
    """获取 query 对应的缓存文件路径"""
    query_hash = get_query_hash(query)
    return os.path.join(CACHE_DIR, f"{query_hash}.json")

def check_cache(query):
    """检查是否存在 query 的缓存文件"""
    file_path = get_cache_file_path(query)
    if os.path.exists(file_path):
        logger.info(f"Cache hit file_path: {file_path} for query: {query}")
        with open(file_path, 'r', encoding='utf-8') as cache_file:
            return json.load(cache_file)  # 返回缓存内容
    return None

 
def save_to_cache(query, data):
    """将数据保存到 query 的缓存文件中"""
    ensure_cache_dir()
    file_path = get_cache_file_path(query)
    with open(file_path, 'w', encoding='utf-8') as cache_file:
        json.dump(data, cache_file, ensure_ascii=False, indent=4)
        logger.info(f"Data cached for query: {query} at {file_path}")

def searx_query(query, top_k):

    
    cached_data = check_cache(query)
    if cached_data:
        logger.info(f"Using cached data for query: {query}")
        return cached_data  # 如果有缓存则直接返回
    # 第一步：模糊查询论文
    logger.info(f"Searching papers for query: {query}")
 
    s = SearxSearchWrapper(searx_host="http://127.0.0.1:8080")
   
    # {
    #     "snippet": result.get("content", ""),
    #     "title": result["title"],
    #     "link": result["url"],
    #     "engines": result["engines"],
    #     "category": result["category"],
    # }
    results = s.results(
        query=query,
        num_results=top_k,

    )
    logger.info(results)
    
    properties_list = [
        
        {
             "chunk_text":  item.get("snippet", ""),
             "ref_id":  get_query_hash(item.get("link", "")),
             "chunk_id":  get_query_hash(item.get("link", "")),
             "title":  item.get("title", ""),
             "link":  item.get("link"),
             "engines":  item.get("engines"),
             "category":  item.get("category"),
        }
         for item in results
    ]
    # 保存 properties_list 到临时文件
    save_to_cache(query, properties_list)

    return properties_list
 


"""Set of constants."""


TYPE_KEY = "__type__"
DATA_KEY = "__data__"
INDEX_STORE_KEY = "index_store"
DOC_STORE_KEY = "doc_store"
DREAMS_ANALYSIS_STORE_KEY = "dreams_analysis_store"
TASK_STEP_STORE_KEY = "task_step_store"


"""analysis store."""

from typing import Dict, Optional, Sequence

from dreamsboard.engine.entity.dreams_personality.dreams_personality import DreamsPersonalityNode
from dreamsboard.engine.generate.code_generate import (
    CodeGenerator
)
from dreamsboard.engine.storage.dreams_analysis_store.types import BaseDreamsAnalysisStore, RefAnalysisInfo
from dreamsboard.engine.storage.dreams_analysis_store.utils import json_to_analysis, analysis_to_json

from dreamsboard.engine.storage.kvstore.types import BaseKVStore

DEFAULT_NAMESPACE = "analysis_store"


class KVDreamsAnalysisStore(BaseDreamsAnalysisStore):
    """analysis (Node) store.

    NOTE: at the moment, this store is primarily used to store Node objects.
    Each node will be assigned an ID.

    The same analysisstore can be reused across index structures. This
    allows you to reuse the same storage for multiple index structures;
    otherwise, each index would create a analysisstore under the hood.

    .. code-block:: python
            store = SimpleDreamsAnalysisStore()
            dreams = DreamsPersonalityNode.from_config(cfg={
                "dreams_guidance_context": "dreams_guidance",
                "dreams_personality_context": "dreams_personality"
            })
            store.add_analysis([dreams])
            logger.info(store.analysis_all)
            dreams_analysis_store_path = concat_dirs(dirname="./storage", basename="dreams_analysis_store.json")
            store.persist(persist_path=dreams_analysis_store_path)

    This will use the same analysisstore for multiple index structures.

    Args:
        kvstore (BaseKVStore): key-value store
        namespace (str): namespace for the analysisstore

    """

    def __init__(
            self,
            kvstore: BaseKVStore,
            namespace: Optional[str] = None,
    ) -> None:
        """Init a KVanalysisStore."""
        self._kvstore = kvstore
        self._namespace = namespace or DEFAULT_NAMESPACE
        self._node_collection = f"{self._namespace}/data"
        self._ref_analysis_collection = f"{self._namespace}/ref_analysis_info"
        self._metadata_collection = f"{self._namespace}/metadata"

    @property
    def analysis_all(self) -> Dict[str, DreamsPersonalityNode]:
        """Get all analysis.

        Returns:
            Dict[str, DreamsPersonalityNode]: analysis

        """
        json_dict = self._kvstore.get_all(collection=self._node_collection)
        return {key: json_to_analysis(json) for key, json in json_dict.items()}

    def add_analysis(
            self, analyses: Sequence[DreamsPersonalityNode], allow_update: bool = True
    ) -> None:
        """Add a analyses to the store.

        Args:
            analyses (List[DreamsPersonalityNode]): analyses
            allow_update (bool): allow update of store from analyses
            :param allow_update:
            :param analyses:

        """
        for node in analyses:
            # NOTE: analysis could already exist in the store, but we overwrite it
            if not allow_update and self.analysis_exists(node.node_id):
                raise ValueError(
                    f"node_id {node.node_id} already exists. "
                    "Set allow_update to True to overwrite."
                )
            node_key = node.node_id
            data = analysis_to_json(node)
            self._kvstore.put(node_key, data, collection=self._node_collection)

            # update analysis_collection if needed
            metadata = {"analysis_hash": node.hash}
            if isinstance(node, DreamsPersonalityNode) and node.ref_analysis_id is not None:
                ref_analysis_info = self.get_ref_analysis_info(node.ref_analysis_id) or RefAnalysisInfo()
                ref_analysis_info.node_ids.append(node.node_id)
                if not ref_analysis_info.metadata:
                    ref_analysis_info.metadata = node.metadata or {}
                self._kvstore.put(
                    node.ref_analysis_id,
                    ref_analysis_info.to_dict(),
                    collection=self._ref_analysis_collection,
                )

                # update metadata with map
                metadata["ref_analysis_id"] = node.ref_analysis_id
                self._kvstore.put(
                    node_key, metadata, collection=self._metadata_collection
                )
            else:
                self._kvstore.put(
                    node_key, metadata, collection=self._metadata_collection
                )

    def get_analysis(self, analysis_id: str, raise_error: bool = True) -> Optional[DreamsPersonalityNode]:
        """Get a analysis from the store.

        Args:
            analysis_id (str): analysis id
            raise_error (bool): raise error if analysis_id not found

        """
        json = self._kvstore.get(analysis_id, collection=self._node_collection)
        if json is None:
            if raise_error:
                raise ValueError(f"analysis_id {analysis_id} not found.")
            else:
                return None
        return json_to_analysis(json)

    def delete_analysis(
            self, analysis_id: str, raise_error: bool = True, remove_ref_analysis_node: bool = True
    ) -> None:
        """Delete a analysis from the store."""
        if remove_ref_analysis_node:
            self._remove_ref_analysis_node(analysis_id)

        delete_success = self._kvstore.delete(analysis_id, collection=self._node_collection)
        _ = self._kvstore.delete(analysis_id, collection=self._metadata_collection)

        if not delete_success and raise_error:
            raise ValueError(f"analysis_id {analysis_id} not found.")

    def analysis_exists(self, analysis_id: str) -> bool:
        """Check if analysis exists."""
        return self._kvstore.get(analysis_id, self._node_collection) is not None

    def get_ref_analysis_info(self, ref_analysis_id: str) -> Optional[RefAnalysisInfo]:
        """Get the RefAnalysisInfo for a given ref_analysis_id."""
        ref_analysis_info = self._kvstore.get(
            ref_analysis_id, collection=self._ref_analysis_collection
        )
        if not ref_analysis_info:
            return None

        # TODO: deprecated legacy support
        if "analysis_ids" in ref_analysis_info:
            ref_analysis_info["node_ids"] = ref_analysis_info.get("analysis_ids", [])
            ref_analysis_info.pop("analysis_ids")

            ref_analysis_info["metadata"] = ref_analysis_info.get("extra_info", {})
            ref_analysis_info.pop("extra_info")

        return RefAnalysisInfo(**ref_analysis_info)

    def get_all_ref_analysis_info(self) -> Optional[Dict[str, RefAnalysisInfo]]:
        """Get a mapping of ref_analysis_id -> RefAnalysisInfo for all ingested nodes."""
        ref_analysis_infos = self._kvstore.get_all(collection=self._ref_analysis_collection)
        if ref_analysis_infos is None:
            return None

        # TODO: deprecated legacy support
        all_ref_analysis_infos = {}
        for analysis_id, ref_analysis_info in ref_analysis_infos.items():
            if "analysis_ids" in ref_analysis_info:
                ref_analysis_info["node_ids"] = ref_analysis_info.get("analysis_ids", [])
                ref_analysis_info.pop("analysis_ids")

                ref_analysis_info["metadata"] = ref_analysis_info.get("extra_info", {})
                ref_analysis_info.pop("extra_info")
                all_ref_analysis_infos[analysis_id] = RefAnalysisInfo(**ref_analysis_info)

        return all_ref_analysis_infos

    def ref_analysis_exists(self, ref_analysis_id: str) -> bool:
        """Check if a ref_analysis_id has been ingested."""
        return self.get_ref_analysis_info(ref_analysis_id) is not None

    def _remove_ref_analysis_node(self, analysis_id: str) -> None:
        """Helper function to remove node analysis_id from ref_analysis_collection."""
        metadata = self._kvstore.get(analysis_id, collection=self._metadata_collection)
        if metadata is None:
            return

        ref_analysis_id = metadata.get("ref_analysis_id", None)

        if ref_analysis_id is None:
            return

        ref_analysis_info = self._kvstore.get(
            ref_analysis_id, collection=self._ref_analysis_collection
        )

        if ref_analysis_info is not None:
            ref_analysis_obj = RefAnalysisInfo(**ref_analysis_info)

            ref_analysis_obj.node_ids.remove(analysis_id)

            # delete ref_analysis from collection if it has no more analysis_ids
            if len(ref_analysis_obj.node_ids) > 0:
                self._kvstore.put(
                    ref_analysis_id,
                    ref_analysis_obj.to_dict(),
                    collection=self._ref_analysis_collection,
                )

            self._kvstore.delete(ref_analysis_id, collection=self._metadata_collection)

    def delete_ref_analysis(self, ref_analysis_id: str, raise_error: bool = True) -> None:
        """Delete a ref_analysis and all it's associated nodes."""
        ref_analysis_info = self.get_ref_analysis_info(ref_analysis_id)
        if ref_analysis_info is None:
            if raise_error:
                raise ValueError(f"ref_analysis_id {ref_analysis_id} not found.")
            else:
                return

        for analysis_id in ref_analysis_info.node_ids:
            self.delete_analysis(analysis_id, raise_error=False, remove_ref_analysis_node=False)

        self._kvstore.delete(ref_analysis_id, collection=self._metadata_collection)
        self._kvstore.delete(ref_analysis_id, collection=self._ref_analysis_collection)



import os

from langchain.schema import Document
from langchain_community.vectorstores.faiss import FAISS
from langchain_community.docstore.in_memory import InMemoryDocstore
 
from dreamsboard.vector.knowledge_base.kb_cache.base import *

from langchain_community.embeddings import HuggingFaceEmbeddings

# patch FAISS to include doc id in Document.metadata
def _new_ds_search(self, search: str) -> Union[str, Document]:
    if search not in self._dict:
        return f"ID {search} not found."
    else:
        doc = self._dict[search]
        if isinstance(doc, Document):
            doc.metadata["id"] = search
        return doc


InMemoryDocstore.search = _new_ds_search


class ThreadSafeFaiss(ThreadSafeObject):
    def __repr__(self) -> str:
        cls = type(self).__name__
        return f"<{cls}: key: {self.key}, obj: {self._obj}, docs_count: {self.docs_count()}>"

    def docs_count(self) -> int:
        return len(self._obj.docstore._dict)

    def save(self, path: str, create_path: bool = True):
        with self.acquire():
            if not os.path.isdir(path) and create_path:
                os.makedirs(path)
            ret = self._obj.save_local(path)
            logger.info(f"已将向量库 {self.key} 保存到磁盘")
        return ret

    def clear(self):
        ret = []
        with self.acquire():
            ids = list(self._obj.docstore._dict.keys())
            if ids:
                ret = self._obj.delete(ids)
                assert len(self._obj.docstore._dict) == 0
            logger.info(f"已将向量库 {self.key} 清空")
        return ret


class _FaissPool(CachePool):
    def new_vector_store(
        self,
        embed_model: str,
        device: str = 'cpu',
    ) -> FAISS:
        # create an empty vector store
        model_kwargs = {'device': device}
        embeddings = HuggingFaceEmbeddings(
            model_name=embed_model,
            model_kwargs=model_kwargs,
            show_progress=True,
        )
        doc = Document(page_content="init", metadata={})
        vector_store = FAISS.from_documents([doc], embeddings, normalize_L2=True)
        ids = list(vector_store.docstore._dict.keys())
        vector_store.delete(ids)
        return vector_store

    def save_vector_store(self, kb_name: str, path: str = None):
        if cache := self.get(kb_name):
            return cache.save(path)

    def unload_vector_store(self, kb_name: str):
        if cache := self.get(kb_name):
            self.pop(kb_name)
            logger.info(f"成功释放向量库：{kb_name}")


class KBFaissPool(_FaissPool):
    def load_vector_store(
        self,
        kb_name: str,
        embed_model: str,
        vector_name: str = None,
        create: bool = True,
        device: str = 'cpu'
    ) -> ThreadSafeFaiss:
        self.atomic.acquire()
        locked = True
        vector_name = vector_name or embed_model.replace(":", "_")
        cache = self.get((kb_name, vector_name))  # 用元组比拼接字符串好一些
        try:
            if cache is None:
                item = ThreadSafeFaiss((kb_name, vector_name), pool=self)
                self.set((kb_name, vector_name), item)
                with item.acquire(msg="初始化"):
                    self.atomic.release()
                    locked = False
                    logger.info(
                        f"loading vector store in '{kb_name}/vector_store/{vector_name}' from disk."
                    )
                    vs_path = os.path.join(kb_name, "vector_store", vector_name) 

                    if os.path.isfile(os.path.join(vs_path, "index.faiss")): 
                        model_kwargs = {'device': device}
                        embeddings = HuggingFaceEmbeddings(
                            model_name=embed_model,
                            model_kwargs=model_kwargs,
                            show_progress=True,
                        )
                        vector_store = FAISS.load_local(
                            vs_path,
                            embeddings,
                            normalize_L2=True,
                            allow_dangerous_deserialization=True,
                        )
                    elif create:
                        # create an empty vector store
                        if not os.path.exists(vs_path):
                            os.makedirs(vs_path)
                        vector_store = self.new_vector_store(
                             embed_model=embed_model, device=device
                        )
                        vector_store.save_local(vs_path)
                    else:
                        raise RuntimeError(f"knowledge base {kb_name} not exist.")
                    item.obj = vector_store
                    item.finish_loading()
            else:
                self.atomic.release()
                locked = False
        except Exception as e:
            if locked:  # we don't know exception raised before or after atomic.release
                self.atomic.release()
            logger.exception(e)
            raise RuntimeError(f"向量库 {kb_name} 加载失败。")
        return self.get((kb_name, vector_name))

 

kb_faiss_pool = KBFaissPool(cache_num=1) 
 


from __future__ import annotations
from langchain_core.messages import ( 
    BaseMessage,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.runnables import Runnable
from langchain.chains import LLMChain
from langchain.chains.ernie_functions import create_structured_output_runnable
from langchain_community.chat_models import ChatOpenAI
from langchain.schema.language_model import BaseLanguageModel
from langchain_core.prompts import ChatPromptTemplate

from dreamsboard.document_loaders.protocol.ner_protocol import Personality


class NerLoader:
    """实体关系抽取"""
    @classmethod
    def form_ner_dreams_personality_builder(cls,
                                            llm_runable: Runnable[LanguageModelInput, BaseMessage]) -> LLMChain:
        """
        生成性格分析的抽取链
        :param llm:
        :return:
        """
        prompt = ChatPromptTemplate.from_messages(
            [
                ("system", """根据您提供的信息，您的性格特点可以总结如下：
        
1. 热情和温柔：您在描述天气和气氛时使用了"温柔长裙风"这样的形容词，表现出您对温暖和舒适的情感。

2. 情感表达：您在文本中表达了对一个叫"宝宝"的角色的期待和关心，这显示了您的感性和情感表达能力。

3. 好奇心和幽默感：您提到了要做大胆的事情，并且以"嘻嘻"结束，这暗示了您对新奇事物的好奇心和幽默感。

4. 关心家人和亲情：您提到了弟弟给了三颗糖，表现出您关心家人的情感。

5. 乐于分享和帮助：您提到要给宝宝剥虾并询问宝宝是否想知道小鱼在说什么，显示出您愿意分享和帮助他人的特点。

6. 可能有一些难以理解的部分：在文本中也出现了一些不太清楚的情节，如呼救情节和提到"小肚小肚"，这可能表现出您的思维有时候会有些混乱或不太连贯。

总的来说，您的性格特点包括热情、情感表达能力、好奇心、幽默感、亲情关怀以及乐于分享和帮助他人。
            
            """),
                ("ai", """personality
热情、情感表达能力、好奇心、幽默感、亲情关怀以及乐于分享和帮助他人"""),
                ("human", "{input}"),
            ]
        )
        chain = create_structured_output_runnable(Personality, llm_runable, prompt)
        return chain



from dreamsboard.engine.constants import DATA_KEY, TYPE_KEY
from dreamsboard.engine.entity.task_step.task_step import TaskStepNode
from dreamsboard.engine.schema import (
    BaseNode,
)


def task_step_to_json(doc: BaseNode) -> dict:
    return {
        DATA_KEY: doc.dict(),
        TYPE_KEY: doc.get_type(),
    }


def json_to_task_step(task_step_dict: dict) -> TaskStepNode:
    task_step_type = task_step_dict[TYPE_KEY]
    data_dict = task_step_dict[DATA_KEY]
    doc: TaskStepNode
 
    if task_step_type == TaskStepNode.get_type():
        doc = TaskStepNode.model_validate(data_dict)
    else:
        raise ValueError(f"Unknown doc type: {task_step_type}")

    return doc

 



from typing import Type, Dict

from dreamsboard.engine.generate.code_generate import (
    CodeGenerator,
    BaseProgramGenerator,
    QueryProgramGenerator,
    AIProgramGenerator,
    EngineProgramGenerator,
)
from dreamsboard.engine.schema import ObjectTemplateType


TEMPLATE_TYPE_TO_GENERATOR_CLASS: Dict[ObjectTemplateType, Type[CodeGenerator]] = {
    ObjectTemplateType.BaseProgramGenerator: BaseProgramGenerator,
    ObjectTemplateType.QueryProgramGenerator: QueryProgramGenerator,
    ObjectTemplateType.AIProgramGenerator: AIProgramGenerator,
    ObjectTemplateType.EngineProgramGenerator: EngineProgramGenerator,
}



"""

""" 

# 03- 故事情境生成 `story_scenario_context`.txt
STORY_BOARD_SCENE_TEMPLATE = """根据文本执行任务。以下任务信息

作为一个社会学研究学者，您已经查阅了《作为激情的爱情》卢曼编写的书籍，尝试通过参考文献中定义的爱情语义学，从文本中总结下方片段
在如下表征信息中，研究其突出特点
研究交流媒介领域的语义信息
研究激情的非理性与风雅情术的偶然性
研究自身的快感是否转移到社会行为上
研究语义信息的固定形式与预期落空因果性，是否存在可激发性拓展到否定物之中
主要突出如下几点
1、我们建议研究社交媒体中的爱情表现。
2、我们提议研究卢曼的爱情语义学在电视和电影中的应用。
3、我们推荐深入研究卢曼的理论，以了解其在社会学研究中的潜在应用。
4、这个研究与社会学研究中的情感因素密切相关。
5、我们的讨论与媒体中的语义信息有关。
6、卢曼的爱情语义学与社交媒体的内容密切相联。
7、我们需要考虑卢曼理论和社会学研究的靠近度。
8、了解卢曼理论如何接近社交媒体中的爱情表现。
9、这个话题与卢曼的爱情语义学及其在社会学中的附近研究领域有关。
10、我们可以探讨卢曼理论和社会学的交叉点。
11、社交媒体中的爱情表现与卢曼理论存在重叠。
12、在这个研究中，卢曼的爱情语义学与社会学观点相交。
你可以尝试分步思考然后告诉我答案，Step by Step Decomposition



文本内容如下

{scene_content}
"""

# 03-故事场景生成 `scene_monologue_context`.txt
STORY_BOARD_SUMMARY_CONTEXT_TEMPLATE = """根据下面人物对话分析这里关于{user_id}今天发生了什么，编写一个独白信息，使用第一人称描述。


人物对话如下

{story_board_summary_context}
"""


# 05-剧情总结 `evolutionary_step`.txt
EDREAMS_EVOLUTIONARY_TEMPLATE = """作为一个社会学研究学者，尝试通过社会学研究相关内容总结下方的对话
社会学研究相关内容分成如下步骤
研究交流媒介领域的语义信息
研究激情的非理性与风雅情术的偶然性
研究自身的快感是否转移到社会行为上
研究语义信息的固定形式与预期落空因果性，是否存在可激发性拓展到否定物之中
你可以尝试分步思考然后告诉我答案，Step by Step Decomposition



人物对话如下

{story_board_summary_context}
"""

# 05-性格分析 `dreams_personality_context`.txt
EDREAMS_PERSONALITY_TEMPLATE = """作为一个社会学研究学者，您已经查阅了《作为激情的爱情》卢曼编写的书籍，尝试通过参考文献中定义的爱情语义学总结下方片段
社会学研究相关内容分成如下步骤
研究交流媒介领域的语义信息
研究激情的非理性与风雅情术的偶然性
研究自身的快感是否转移到社会行为上
研究语义信息的固定形式与预期落空因果性，是否存在可激发性拓展到否定物之中

根据上述信息,总结片段中人物的性格
你可以尝试分步思考然后告诉我答案，Step by Step Decomposition



{evolutionary_step}
"""


# 04-情感情景引导.txt
DREAMS_GEN_TEMPLATE = """作为一个心理咨询工作者，尝试使用开放性问题与来访者交流，尝试带入他们的感受和思维，下面内容中已经整理了关于{user_id}的`故事梗概`，`故事情境`,`故事场景`，请告诉我如何引导来访者说出自己的问题
你可以尝试分步思考然后告诉我答案，Step by Step Decomposition

例如
**Step 1: 建立情感连接**
开始时，您可以通过表达理解和共鸣来建立情感连接，让来访者感到舒适。您可以说：“我注意到这个对话中有许多愉快的时刻和互动。你对这些时刻有什么特别的感受吗？”

**Step 2: 探索来访者的感受**
继续引导来访者表达他们的感受。您可以问：“在这个对话中，有哪些瞬间让你感到开心或快乐？”

**Step 3: 询问是否有反感情绪**
除了积极的情感，也要问询是否有一些负面情感或担忧。您可以说：“除了快乐的瞬间，是否有一些让你感到不安或担忧的地方？”



# 故事梗概
 
{story_board_summary_context}


# 故事情境

{story_scenario_context}


# 故事场景

{scene_monologue_context}
"""




"""

"""

# 00-判断情感表征是否符合.txt
AEMO_REPRESENTATION_PROMPT_TEMPLATE = """作为一个社会学研究学者，您已经查阅了《作为激情的爱情》卢曼编写的书籍，尝试通过参考文献中定义的爱情语义学，总结下方片段
社会学研究相关内容分成如下步骤
研究交流媒介领域的语义信息
研究激情的非理性与风雅情术的偶然性
研究自身的快感是否转移到社会行为上
研究语义信息的固定形式与预期落空因果性，是否存在可激发性拓展到否定物之中
你可以尝试分步思考然后告诉我答案，Step by Step Decomposition



{start_task_context}
"""
 


from langchain_core.pydantic_v1 import BaseModel, Field


class Personality(BaseModel):
    """性格信息."""

    personality: str = Field(..., description="性格评价")


class DreamsStepInfo(BaseModel):
    """开放性引导问题"""
    step_advice: str = Field(..., description="""Advice provided in this step, e.g. "I would say something like: 'I understand this is a difficult situation for you.'""")
    step_description: str = Field(..., description="""(Description of the counseling step, e.g. "Establish trust" """)


class DreamsStepInfoListWrapper(BaseModel):
    """开放问题询问步骤列表."""
    steps: list[DreamsStepInfo] = Field(..., description="""List of steps in the counseling process""")

class TaskStepNode(BaseModel):
    """任务步骤节点."""
    start_task_context: str = Field(..., description="""开始任务""")
    aemo_representation_context: str = Field(..., description=""" 任务总体描述""")
    task_step_name: str = Field(..., description="""提取步骤的名称，例如"分析近几年研究领域的技术框架与方法论"、"基于规则的方法"、"研究论文中采用的主要框架"、"Transformer"等""")
    task_step_description: str = Field(..., description="""提取每个步骤的具体建议和问题，例如"Text2SQL 是将自然语言查询（NLQ）转换为结构化查询语言（SQL）的任务，近年来在数据库和自然语言处理（NLP）领域受到广泛关注。主要技术框架和方法论包括："等""")
    task_step_level: str = Field(..., description="""提取步骤的层级编号，只有root>child层级，例如"0"、"0>1"、"0>2"、"1"、"1>1"、"1>2"等""")
    
class TaskStepNodeListWrapper(BaseModel):
    """任务步骤列表."""
    steps: list[TaskStepNode] = Field(..., description="""List of steps in the counseling process""")

    def to_dict(self):
        return [step.to_dict() for step in self.steps]


class TaskStepRefineNode(BaseModel):
    """根据批评意见优化当前回答并续写上下文内容."""
    thought: str = Field(..., description="""优化后的回答""")
    answer: str = Field(..., description=""" 续写的上下文内容""")
    answer_socre: str = Field(..., description="""答案评分分数。 浮点数""")
    



TASK_MD_TEMPLATE = """
# {start_task_context} 


{context_placeholder}
 
"""



TASK_STEP_MD_TITLE_TEMPLATE = """
{task_step_name} [task_id:{task_step_level}]({task_step_id})

{task_step_question_answer}
"""


TASK_STEP_MD_DESC_TEMPLATE = """
{task_step_name} [task_id:{task_step_level}]({task_step_id}) {task_step_question_answer}
"""



TASK_STEP_MD_LIST_TEMPLATE = """
- **{task_step_name} [task_id:{task_step_level}]({task_step_id})**:{task_step_question_answer}
"""


TASK_STEP_MD_TEMPLATE = """
{task_step_name} [task_id:{task_step_level}]({task_step_id}), {task_step_question_answer}
"""



 
from multiprocessing import Process, Queue, Lock

from threading import Thread
import traceback
import threading 
import logging 
import time
import queue
import asyncio
import atexit



logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)

logger.addHandler(handler)



class Iteratorize:
    """
    这个类会将一个任务封装为延迟执行的迭代器（Generator）。
    任务不会在注册时立即启动线程，而是在 __iter__ 方法第一次被调用时才会启动线程。
    每个任务会检查资源锁，确保同一资源不会被多个任务同时访问。
    """
    def __init__(self, func, resource_id, sentinel, kwargs={}, resource_lock=None):
        self.mfunc = func
        self.q = Queue()
        self.sentinel = sentinel
        self.kwargs = kwargs
        self.stop_now = False
        self.resource_id = resource_id
        self.resource_lock = resource_lock  # 传入具体的锁，而不是整个lock_dict

        self._callback = self._callback_func()  # 保存回调函数
        self.thread = Thread(target=self._run)  # 延迟创建线程
        self.thread.start()

    def _callback_func(self):
        """回调函数用于收集结果，并控制迭代的结束"""
        def _callback(val):
            if self.stop_now:
                raise ValueError
            self.q.put(val)
        return _callback
    
    def _run(self):
        """在迭代开始时启动线程并执行任务
        DO：在 EventManager 中，多个事件（task_function）可能会同时尝试获取相同资源的锁（resource_lock），
        但如果一个线程持有某个锁且另一个线程正在等待同一个锁，它们就会产生死锁。
        如果多个任务被注册到相同资源，并且这些任务还相互依赖（或者访问同一个资源），则可能会形成死锁链。
        解决方案: 对资源锁增加一个超时机制，若在一定时间内无法获得锁，可以进行回滚，避免无限期阻塞。
        """
   
        owner = f"getlock thread {threading.get_native_id()}"
        logger.info(f"owner:{owner}, resource_id:{self.resource_id}")
         
        self.resource_lock.acquire()
        owner = f"lock thread {threading.get_native_id()}"
        logger.info(f"owner:{owner}, resource_id:{self.resource_id}")
    
        try:
            # 执行任务，传入回调函数
            self.mfunc(callback=self._callback, resource_id=self.resource_id, **self.kwargs)
        except ValueError:
            traceback.print_exc()
        except:
            traceback.print_exc()
        finally:
            
            self.resource_lock.release()
            # 任务结束，放入sentinel来标识结束
            self.q.put(self.sentinel)
     

    def __iter__(self):
        """控制迭代器，第一次迭代时启动线程"""
        return self


    def __next__(self):
        """ Fetch the next value from the queue or stop if sentinel is received. """
        obj = self.q.get(True, None)
        if obj == self.sentinel:  # 使用 == 比较
            raise StopIteration
        else:
            return obj 

    def __del__(self):
        """ Cleanup resources, stop the task if needed. """ 
        self.stop_now = True
        if self.thread.is_alive():
            self.thread.join()

    def __enter__(self):
        """ Used for context management. """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Stop the iteration when exiting context. """
        self.stop_now = True
        if self.thread.is_alive():
            self.thread.join()



class ProcessIteratorize:
    """
    这个类会将一个任务封装为延迟执行的迭代器（Generator）。
    任务不会在注册时立即启动线程，而是在 __iter__ 方法第一次被调用时才会启动线程。
    每个任务会检查资源锁，确保同一资源不会被多个任务同时访问。
    """
    def __init__(self, func, resource_id, sentinel, kwargs={}, resource_lock=None):
        self.mfunc = func
        self.q = Queue()
        self.sentinel = sentinel
        self.kwargs = kwargs
        self.stop_now = False
        self.resource_id = resource_id
        self.resource_lock = resource_lock  # 传入具体的锁，而不是整个lock_dict

        self._callback = self._callback_func()  # 保存回调函数
        self.process = Process(target=self._run)  # 延迟创建进程
        self.process.start()
        atexit.register(self.cleanup)  # Jupyter 退出时清理进程

    def _callback_func(self):
        """回调函数用于收集结果，并控制迭代的结束"""
        def _callback(val):
            if self.stop_now:
                raise ValueError
            self.q.put(val)
        return _callback
    
    def _run(self):
        """在迭代开始时启动线程并执行任务
        DO：在 EventManager 中，多个事件（task_function）可能会同时尝试获取相同资源的锁（resource_lock），
        但如果一个线程持有某个锁且另一个线程正在等待同一个锁，它们就会产生死锁。
        如果多个任务被注册到相同资源，并且这些任务还相互依赖（或者访问同一个资源），则可能会形成死锁链。
        解决方案: 对资源锁增加一个超时机制，若在一定时间内无法获得锁，可以进行回滚，避免无限期阻塞。
        """
   
        owner = f"getlock thread {threading.get_native_id()}"
        logger.info(f"owner:{owner}, resource_id:{self.resource_id}")
         
        self.resource_lock.acquire()
        owner = f"lock thread {threading.get_native_id()}"
        logger.info(f"owner:{owner}, resource_id:{self.resource_id}")
    
        try:
            # 执行任务，传入回调函数
            self.mfunc(callback=self._callback, resource_id=self.resource_id, **self.kwargs)
        except ValueError:
            traceback.print_exc()
        except:
            traceback.print_exc()
        finally:
            
            self.resource_lock.release()
            # 任务结束，放入sentinel来标识结束
            self.q.put(self.sentinel)
     

    def __iter__(self):
        """控制迭代器，第一次迭代时启动线程"""
        return self


    def __next__(self):
        """ Fetch the next value from the queue or stop if sentinel is received. """
        obj = self.q.get(True, None)
        if obj == self.sentinel:  # 使用 == 比较
            raise StopIteration
        else:
            return obj

    def cleanup(self):
        self.stop_now = True
        if self.process.is_alive():
            self.process.terminate()
            self.process.join()

    def __del__(self):
        """ Cleanup resources, stop the task if needed. """ 
        pass 

    def __enter__(self):
        """ Used for context management. """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Stop the iteration when exiting context. """
        self.stop_now = True
        if self.process.is_alive():
            self.process.terminate()
            self.process.join()


def call_func(func, resource_id,  kwargs={}):
    # 创建一个资源锁，用于保护资源
    resource_lock = Lock()
    sentinel = "SENTINEL"
    result_holder = []

    # 创建多个 Iteratorize 实例，模拟多个并发任务
    it = ProcessIteratorize(func=func, resource_id=resource_id, sentinel=sentinel, resource_lock=resource_lock, kwargs=kwargs)

    thread = threading.Thread(target=run_iterator, args=(it,result_holder))
   
    thread.start()
    thread.join()
    
    return result_holder
 

# 运行 Iteratorize 并获取结果
def run_iterator(it, result_holder):
    with it as iterator:
        for result in iterator:
            print(f"Received: {result}")
            result_holder.append(result)
        print(f"Iteration complete for {it.resource_id}.")


class EventManager:
    def __init__(self):
        self.lock_dict = {}  # 用来存储资源ID对应的锁
        self.tasks = queue.Queue()  # 使用Queue来存储事件任务
        self.results = {}  # 使用字典存储每个eventId对应的结果
        self.lock = threading.Lock()  # 添加锁来同步对lock_dict的访问
        self.event_counter = 0  # 用于生成唯一的eventId
        self.event_counter_lock = threading.RLock()  # 用于保护event_counter的原子递增操作
        self.execution_thread = threading.Thread(target=self._execute_task_queue, daemon=True)  # 使用 daemon 线程
        self.execution_thread.start()  # 启动任务执行线程

    def register_event(self, task_func, resource_id, kwargs=None):
        """ 注册事件并绑定资源ID，如果资源ID对应的锁已被占用，等待直到资源释放 """
        kwargs = kwargs or {}

        # 使用锁来同步对lock_dict的访问
        try:
            self.lock.acquire()
            if resource_id not in self.lock_dict:
                self.lock_dict[resource_id] = threading.Lock()  # 为每个资源ID创建一个锁
        finally:
            self.lock.release()

        # 生成唯一的eventId
        event_id = self.generate_event_id(resource_id) 
        sentinel = "SENTINEL"
        # 创建Iteratorize任务并传入对应的资源锁
        iteratorize_task = Iteratorize(task_func, resource_id,  sentinel=sentinel, resource_lock=self.lock_dict[resource_id], kwargs=kwargs)
        
        # 将任务存储到队列中
        self.tasks.put((event_id, iteratorize_task))  # 将event_id和task作为元组放入队列中

        # 为该eventId初始化结果存储
        self.results[event_id] = []

        return event_id
 
    def _execute_task_queue(self):
        """ 使用asyncio异步处理任务队列 """
 
        while True:
            # 从队列中取任务，如果队列为空则会阻塞，直到有新的任务加入
            event_id, task = self.tasks.get()

            if task:  
                for result in task:
                    self.results[event_id].append(result)
                 
    def get_results(self, event_id=None):
        """ 获取某个任务的执行结果，如果没有传入event_id，返回所有任务的结果 """
        if event_id in self.results:
            return self.results[event_id]
        else:
            return self.results

    def generate_event_id(self, resource_id):
        """生成唯一的事件ID，格式：event:{resource_id}:{event_counter}"""
        try:
            self.event_counter_lock.acquire()
            self.event_counter += 1
            return f"event:{resource_id}:{self.event_counter}"
        finally:
            self.event_counter_lock.release()

    def clean_up_all_tasks(self):
        """ 清理所有已完成的任务包括资源 """
        self.results.clear()
event_manager = EventManager()


from typing import List

from llama_index.core import (
    SimpleDirectoryReader,
    VectorStoreIndex,
    StorageContext,
    load_index_from_storage, Document,
)

from llama_index.core.tools import QueryEngineTool, ToolMetadata
from llama_index.embeddings.openai import OpenAIEmbedding
from llama_index.llms.openai import OpenAI


def build_query_docs() -> list[list[Document]]:
    # load data

    # load data
    meas_docs = SimpleDirectoryReader(
        input_files=["/home/dmeck/Documents/电网业务中台/电网资源业务中台接口开发规范说明书V2.01-测点管理中心.docx"]
    ).load_data()
    topo_docs = SimpleDirectoryReader(
        input_files=["/home/dmeck/Documents/电网业务中台/电网资源业务中台接口开发规范说明书V2.01-电网拓扑中心.docx"]
    ).load_data()
    geo_docs = SimpleDirectoryReader(
        input_files=["/home/dmeck/Documents/电网业务中台/电网资源业务中台接口开发规范说明书V2.01-电网图形中心.docx"]
    ).load_data()
    asset_docs = SimpleDirectoryReader(
        input_files=["/home/dmeck/Documents/电网业务中台/电网资源业务中台接口开发规范说明书V2.01-电网资产中心.docx"]
    ).load_data()
    psr_docs = SimpleDirectoryReader(
        input_files=["/home/dmeck/Documents/电网业务中台/电网资源业务中台接口开发规范说明书V2.01-电网资源中心.docx"]
    ).load_data()
    base_docs = SimpleDirectoryReader(
        input_files=["/home/dmeck/Documents/电网业务中台/电网资源业务中台接口开发规范说明书V2.01-基础服务中心.docx"]
    ).load_data()
    return [meas_docs, topo_docs, geo_docs, asset_docs, psr_docs, base_docs]


def build_query_docs_index_store(query_docs: list[list[Document]], embeddings: OpenAIEmbedding) -> list[VectorStoreIndex]:
    try:
        storage_context = StorageContext.from_defaults(
            persist_dir="./storage/meas"
        )
        meas_index = load_index_from_storage(storage_context)
        storage_context = StorageContext.from_defaults(
            persist_dir="./storage/topo"
        )
        topo_index = load_index_from_storage(storage_context)

        storage_context = StorageContext.from_defaults(
            persist_dir="./storage/geo"
        )
        geo_index = load_index_from_storage(storage_context)
        storage_context = StorageContext.from_defaults(
            persist_dir="./storage/asset"
        )
        asset_index = load_index_from_storage(storage_context)
        storage_context = StorageContext.from_defaults(
            persist_dir="./storage/psr"
        )
        psr_index = load_index_from_storage(storage_context)
        storage_context = StorageContext.from_defaults(
            persist_dir="./storage/base"
        )
        base_index = load_index_from_storage(storage_context)
        index_loaded = True
    except:
        index_loaded = False
    if not index_loaded:
        # load data
        meas_docs = query_docs[0]
        topo_docs = query_docs[1]
        geo_docs = query_docs[2]
        asset_docs = query_docs[3]
        psr_docs = query_docs[4]
        base_docs = query_docs[5]

        # build index
        meas_index = VectorStoreIndex.from_documents(
            meas_docs,embed_model=embeddings
        )
        topo_index = VectorStoreIndex.from_documents(
            topo_docs,embed_model=embeddings
        )
        geo_index = VectorStoreIndex.from_documents(
            geo_docs,embed_model=embeddings
        )
        asset_index = VectorStoreIndex.from_documents(
            asset_docs,embed_model=embeddings
        )
        psr_index = VectorStoreIndex.from_documents(
            psr_docs,embed_model=embeddings
        )
        base_index = VectorStoreIndex.from_documents(
            base_docs,embed_model=embeddings
        )

        # persist index
        meas_index.storage_context.persist(persist_dir="./storage/meas")
        topo_index.storage_context.persist(persist_dir="./storage/topo")
        geo_index.storage_context.persist(persist_dir="./storage/geo")
        asset_index.storage_context.persist(persist_dir="./storage/asset")
        psr_index.storage_context.persist(persist_dir="./storage/psr")
        base_index.storage_context.persist(persist_dir="./storage/base")

    index_store = [meas_index, topo_index, geo_index, asset_index, psr_index, base_index]
    return index_store


def build_query_engine_tools(llm: OpenAI, index_store: list[VectorStoreIndex]) -> list[QueryEngineTool]:
    meas_index = index_store[0]
    topo_index = index_store[1]
    geo_index = index_store[2]
    asset_index = index_store[3]
    psr_index = index_store[4]
    base_index = index_store[5]
    meas_engine = meas_index.as_query_engine(similarity_top_k=3, llm=llm)
    topo_engine = topo_index.as_query_engine(similarity_top_k=3, llm=llm)
    geo_engine = geo_index.as_query_engine(similarity_top_k=3, llm=llm)
    asset_engine = asset_index.as_query_engine(similarity_top_k=3, llm=llm)
    psr_engine = psr_index.as_query_engine(similarity_top_k=3, llm=llm)
    base_engine = base_index.as_query_engine(similarity_top_k=3, llm=llm)
    query_tool_meas = QueryEngineTool.from_defaults(
        query_engine=meas_engine,
        name="meas",
        description=(
            f"关于测点管理中心的业务中台，包括了测量查询服务、事件中心服务、事件类型定义、事件代码表、策略类型、实时开源状态 等业务信息"
            f"eg:检索时带上详细的问题内容"

        ),
    )

    query_tool_topo = QueryEngineTool.from_defaults(
        query_engine=topo_engine,
        name="topo",
        description=(
            f"关于电网拓扑中心的业务中台，包括了一、电网拓扑中心概述、二、拓扑基础分析服务群、三、拓扑高级分析服务 等业务信息"
            f"eg:检索时带上详细的问题内容"

        ),
    )
    query_tool_geo= QueryEngineTool.from_defaults(
        query_engine=geo_engine,
        name="geo",
        description=(
            f"关于电网图形中心的业务中台，包括了一、空间查询服务、二、专题图成图服务、三、Gis出图服务、专题图出图服务 等业务信息"
            f"eg:检索时带上详细的问题内容"

        ),
    )
    query_tool_asset = QueryEngineTool.from_defaults(
        query_engine=asset_engine,
        name="asset",
        description=(
            f"关于电网资产中心的业务中台，包括了资产信息查询服务 业务信息"
            f"eg:检索时带上详细的问题内容"

        ),
    )

    query_tool_psr = QueryEngineTool.from_defaults(
        query_engine=psr_engine,
        name="psr",
        description=(
            f"关于电网资源中心的业务中台，包括了资源查询服务 业务信息"
            f"eg:检索时带上详细的问题内容"

        ),
    )
    query_tool_base = QueryEngineTool.from_defaults(
        query_engine=base_engine,
        name="base",
        description=(
            f"关于电基础服务中心的业务中台，包括了3.1电网变更服务、3.2统一认证服务、3.3通用查询服务、3.4结构树服务、3.5权限服务、3.6人员组织查询服务 业务信息"
            f"eg:检索时带上详细的问题内容"

        ),
    )
    query_engine_tools = [query_tool_meas, query_tool_topo, query_tool_geo, query_tool_asset, query_tool_psr, query_tool_base]
    return query_engine_tools



import json


# 解析愿jsonl文件
def parse_jsonl_file(jsonl_file):
    with open(jsonl_file, "r", encoding='utf-8') as f:
        for line in f:
            yield json.loads(line)


def main():
    #     获取jsonl文件
    jsonl_file = "gemini_q_glm_a_finetuning_events_1152q_1710191088.258371.jsonl"
    jsonl_data = parse_jsonl_file(jsonl_file)

    # 转换为json格式
    json_data = []
    for line in jsonl_data:
        messages = line["messages"]
        # 获取系统消息, 作为system字段，并从messages中删除
        for message in messages:
            if message['role'] == "system":
                messages.remove(message)
                line['system'] = message['content']
                break
        json_data.append(line)

    # 写入json文件
    with open("gemini_q_glm_a_finetuning_events_1152q_1710191088.258371.json", "w") as f:
        json.dump(json_data, f, indent=4, ensure_ascii=False)


def print_json():
    #     获取jsonl文件
    jsonl_file = "/home/dmeck/Documents/content/finetuning_events_10q_2023-11-30_170732.jsonl"
    jsonl_data = parse_jsonl_file(jsonl_file)

    # 转换为json格式
    json_data = []
    for line in jsonl_data:
        messages = line["messages"]
        for message in messages:
            if message['role'] == "system":
                print('system'+message['content'])
            if message['role'] == "assistant":
                print('assistant'+message['content'])


if __name__ == "__main__":
    main()



import os
import shutil
from typing import Dict, List, Tuple, Any, Sized

from langchain.docstore.document import Document
 
from dreamsboard.vector.knowledge_base.kb_cache.faiss_cache import (
    ThreadSafeFaiss,
    kb_faiss_pool,
) 
from dreamsboard.vector.base import CollectionService, DocumentWithVSId

def _len_check_if_sized(x: Any, y: Any, x_name: str, y_name: str) -> None:
    if isinstance(x, Sized) and isinstance(y, Sized) and len(x) != len(y):
        raise ValueError(
            f"{x_name} and {y_name} expected to be equal length but "
            f"len({x_name})={len(x)} and len({y_name})={len(y)}"
        )
    return

class FaissCollectionService(CollectionService):
    _vs_path: str
    device: str
    kb_name: str
    embed_model: str
    vector_name: str = None

    def __init__(self, kb_name: str, embed_model: str, vector_name: str, device: str = 'cpu') -> None:
        self.device = device
        self.kb_name = kb_name
        self.embed_model = embed_model
        self.vector_name = vector_name
        self._vs_path = os.path.join(kb_name, "vector_store", vector_name)
        self.load_vector_store()
 
    def load_vector_store(self) -> ThreadSafeFaiss:
        return kb_faiss_pool.load_vector_store(
            kb_name=self.kb_name,
            vector_name=self.vector_name,
            embed_model=self.embed_model,
            device=self.device,
        )

    def save_vector_store(self):
        self.load_vector_store().save(self._vs_path)

    def get_doc_by_ids(self, ids: List[str]) -> List[DocumentWithVSId]:
        if not isinstance(ids[0], str):
            raise ValueError(f"ids expected to be List[str] but got {type(ids)}")

        with self.load_vector_store().acquire() as vs:
            docs = []
            for id in ids:
                store_data = vs.docstore._dict.get(id)
                if store_data is None:
                    continue
                doc = DocumentWithVSId(**{**store_data.dict(), "id":id})
                docs.append(doc)
            return docs

    def del_doc_by_ids(self, ids: List[str]) -> bool:
        with self.load_vector_store().acquire() as vs:
            vs.delete(ids)
 

    def do_create_kb(self):
        if not os.path.exists(self._vs_path):
            os.makedirs(self._vs_path)
        self.load_vector_store()
 

    def do_search(
        self,
        query: str,
        top_k: int,
        score_threshold: float = 1,
    ) -> List[DocumentWithVSId]:
        with self.load_vector_store().acquire(msg="查询") as vs:
            faiss_retriever = vs.as_retriever(
                search_type="similarity_score_threshold",
                search_kwargs={"score_threshold": score_threshold, "k": top_k},
            )
            docs = faiss_retriever.invoke(query)
            docs = [DocumentWithVSId(**{**doc.dict(), "id":doc.metadata['id']}) for doc in docs]
        return docs

    def do_add_doc(
        self,
        docs: List[DocumentWithVSId],
        **kwargs,
    ) -> List[Dict]:
        if docs is None or len(docs) == 0:
            return []
        if not isinstance(docs[0], DocumentWithVSId):
            raise ValueError(f"docs expected to be List[DocumentWithVSId] but got {type(docs)}")

        ids = [x.id for x in docs]
        texts = [x.page_content for x in docs]
        metadatas = [x.metadata for x in docs]


        _len_check_if_sized(docs, ids, "docs", "ids")
        _len_check_if_sized(docs, texts, "docs", "texts")
        _len_check_if_sized(docs, metadatas, "docs", "metadatas")
        with self.load_vector_store().acquire(msg="插入") as vs:
            embeddings = vs.embeddings.embed_documents(texts)
            ids = vs.add_embeddings(
                text_embeddings=zip(texts, embeddings), metadatas=metadatas, ids=ids
            ) 

        self.save_vector_store()
        doc_infos = [DocumentWithVSId(**{**doc.dict(), "id":id}) for id, doc in zip(ids, docs)]
        return doc_infos


    def do_clear_vs(self):
        with kb_faiss_pool.atomic:
            kb_faiss_pool.pop((self.kb_name, self.vector_name))
        try:
            shutil.rmtree(self._vs_path)
        except Exception:
            ...
        os.makedirs(self._vs_path, exist_ok=True)

 

 



"""analysis store."""

from typing import Dict, Optional, Sequence

from dreamsboard.engine.entity.task_step.task_step import TaskStepNode
from dreamsboard.engine.generate.code_generate import (
    CodeGenerator
)
from dreamsboard.engine.storage.task_step_store.types import BaseTaskStepStore, RefTaskStepInfo
from dreamsboard.engine.storage.task_step_store.utils import json_to_task_step, task_step_to_json

from dreamsboard.engine.storage.kvstore.types import BaseKVStore

DEFAULT_NAMESPACE = "task_step_store"


class KVTaskStepStore(BaseTaskStepStore):
    """task_step (Node) store.

    NOTE: at the moment, this store is primarily used to store TaskStepNode objects.
    Each node will be assigned an ID.

    The same task_stepstore can be reused across index structures. This
    allows you to reuse the same storage for multiple index structures;
    otherwise, each index would create a task_stepstore under the hood.

    .. code-block:: python
            store = SimpleTaskStepStore()
            task_step = TaskStepNode.from_config(cfg={
                "task_step_guidance_context": "task_step_guidance",
                "task_step_personality_context": "task_step_personality"
            })
            store.add_task_step([task_step])
            logger.info(store.task_step_all)
            task_step_store_path = concat_dirs(dirname="./storage", basename="task_step_store.json")
            store.persist(persist_path=task_step_store_path)

    This will use the same task_stepstore for multiple index structures.

    Args:
        kvstore (BaseKVStore): key-value store
        namespace (str): namespace for the task_stepstore

    """

    def __init__(
            self,
            kvstore: BaseKVStore,
            namespace: Optional[str] = None,
    ) -> None:
        """Init a KVtask_stepStore."""
        self._kvstore = kvstore
        self._namespace = namespace or DEFAULT_NAMESPACE
        self._node_collection = f"{self._namespace}/data"
        self._ref_task_step_collection = f"{self._namespace}/ref_task_step_info"
        self._metadata_collection = f"{self._namespace}/metadata"

    @property
    def task_step_all(self) -> Dict[str, TaskStepNode]:
        """Get all task_step.

        Returns:
            Dict[str, TaskStepNode]: task_step

        """
        json_dict = self._kvstore.get_all(collection=self._node_collection)
        return {key: json_to_task_step(json) for key, json in json_dict.items()}

    def add_task_step(
            self, task_steps: Sequence[TaskStepNode], allow_update: bool = True
    ) -> None:
        """Add a task_steps to the store.

        Args:
            task_steps (List[TaskStepNode]): task_steps
            allow_update (bool): allow update of store from task_steps
            :param allow_update:
            :param task_steps:

        """
        for node in task_steps:
            # NOTE: task_step could already exist in the store, but we overwrite it
            if not allow_update and self.task_step_exists(node.node_id):
                raise ValueError(
                    f"node_id {node.node_id} already exists. "
                    "Set allow_update to True to overwrite."
                )
            node_key = node.node_id
            data = task_step_to_json(node)
            self._kvstore.put(node_key, data, collection=self._node_collection)

            # update analysis_collection if needed
            metadata = {"task_step_hash": node.hash}
            if isinstance(node, TaskStepNode) and node.ref_task_step_id is not None:
                ref_task_step_info = self.get_ref_task_step_info(node.ref_task_step_id) or RefTaskStepInfo()
                ref_task_step_info.node_ids.append(node.node_id)
                if not ref_task_step_info.metadata:
                    ref_task_step_info.metadata = node.metadata or {}
                self._kvstore.put(
                    node.ref_task_step_id,
                    ref_task_step_info.to_dict(),
                    collection=self._ref_task_step_collection,
                )

                # update metadata with map
                metadata["ref_task_step_id"] = node.ref_task_step_id
                self._kvstore.put(
                    node_key, metadata, collection=self._metadata_collection
                )
            else:
                self._kvstore.put(
                    node_key, metadata, collection=self._metadata_collection
                )

    def get_task_step(self, task_step_id: str, raise_error: bool = True) -> Optional[TaskStepNode]:
        """Get a task_step from the store.

        Args:
            task_step_id (str): task_step id
            raise_error (bool): raise error if task_step_id not found

        """
        json = self._kvstore.get(task_step_id, collection=self._node_collection)
        if json is None:
            if raise_error:
                raise ValueError(f"task_step_id {task_step_id} not found.")
            else:
                return None
        return json_to_task_step(json)

    def delete_task_step(
            self, task_step_id: str, raise_error: bool = True, remove_ref_task_step_node: bool = True
    ) -> None:
        """Delete a task_step from the store."""
        if remove_ref_task_step_node:
            self._remove_ref_task_step_node(task_step_id)

        delete_success = self._kvstore.delete(task_step_id, collection=self._node_collection)
        _ = self._kvstore.delete(task_step_id, collection=self._metadata_collection)

        if not delete_success and raise_error:
            raise ValueError(f"task_step_id {task_step_id} not found.")

    def task_step_exists(self, task_step_id: str) -> bool:
        """Check if task_step exists."""
        return self._kvstore.get(task_step_id, collection=self._node_collection) is not None

    def get_ref_task_step_info(self, ref_task_step_id: str) -> Optional[RefTaskStepInfo]:
        """Get the RefTaskStepInfo for a given ref_task_step_id."""
        ref_task_step_info = self._kvstore.get(
            ref_task_step_id, collection=self._ref_task_step_collection
        )
        if not ref_task_step_info:
            return None

        # TODO: deprecated legacy support
        if "task_step_ids" in ref_task_step_info:
            ref_task_step_info["node_ids"] = ref_task_step_info.get("task_step_ids", [])
            ref_task_step_info.pop("task_step_ids")

            ref_task_step_info["metadata"] = ref_task_step_info.get("extra_info", {})
            ref_task_step_info.pop("extra_info")

        return RefTaskStepInfo(**ref_task_step_info)

    def get_all_ref_task_step_info(self) -> Optional[Dict[str, RefTaskStepInfo]]:
        """Get a mapping of ref_task_step_id -> RefTaskStepInfo for all ingested nodes."""
        ref_task_step_infos = self._kvstore.get_all(collection=self._ref_task_step_collection)
        if ref_task_step_infos is None:
            return None

        # TODO: deprecated legacy support
        all_ref_task_step_infos = {}
        for task_step_id, ref_task_step_info in ref_task_step_infos.items():
            if "task_step_ids" in ref_task_step_info:
                ref_task_step_info["node_ids"] = ref_task_step_info.get("task_step_ids", [])
                ref_task_step_info.pop("task_step_ids")

                ref_task_step_info["metadata"] = ref_task_step_info.get("extra_info", {})
                ref_task_step_info.pop("extra_info")
                all_ref_task_step_infos[task_step_id] = RefTaskStepInfo(**ref_task_step_info)

        return all_ref_task_step_infos

    def ref_task_step_exists(self, ref_task_step_id: str) -> bool:
        """Check if a ref_task_step_id has been ingested."""
        return self.get_ref_task_step_info(ref_task_step_id) is not None

    def _remove_ref_task_step_node(self, task_step_id: str) -> None:
        """Helper function to remove node task_step_id from ref_task_step_collection."""
        metadata = self._kvstore.get(task_step_id, collection=self._metadata_collection)
        if metadata is None:
            return

        ref_task_step_id = metadata.get("ref_task_step_id", None)

        if ref_task_step_id is None:
            return

        ref_task_step_info = self._kvstore.get(
            ref_task_step_id, collection=self._ref_task_step_collection
        )

        if ref_task_step_info is not None:
            ref_task_step_obj = RefTaskStepInfo(**ref_task_step_info)

            ref_task_step_obj.node_ids.remove(task_step_id)

            # delete ref_task_step from collection if it has no more task_step_ids
            if len(ref_task_step_obj.node_ids) > 0:
                self._kvstore.put(
                    ref_task_step_id,
                    ref_task_step_obj.to_dict(),
                    collection=self._ref_task_step_collection,
                )

            self._kvstore.delete(ref_task_step_id, collection=self._metadata_collection)

    def delete_ref_task_step(self, ref_task_step_id: str, raise_error: bool = True) -> None:
        """Delete a ref_task_step and all it's associated nodes."""
        ref_task_step_info = self.get_ref_task_step_info(ref_task_step_id)
        if ref_task_step_info is None:
            if raise_error:
                raise ValueError(f"ref_task_step_id {ref_task_step_id} not found.")
            else:
                return

        for task_step_id in ref_task_step_info.node_ids:
            self.delete_task_step(task_step_id, raise_error=False, remove_ref_task_step_node=False)

        self._kvstore.delete(ref_task_step_id, collection=self._metadata_collection)
        self._kvstore.delete(ref_task_step_id, collection=self._ref_task_step_collection)



def save_questions(questions, path):
    with open(path, "w", encoding='utf-8') as f:
        for question in questions:
            f.write(question + "\n")


def load_questions(path):
    questions = []
    with open(path, "r") as f:
        for line in f:
            questions.append(line.strip())
    return questions



import os
import shutil
from typing import Dict, List, Tuple

from langchain.docstore.document import Document
 
class DocumentWithVSId(Document):
    """
    矢量化后的文档
    """
    id: str = None


class CollectionService:
    _vs_path: str
    device: str
    kb_name: str
    embed_model: str
    vector_name: str = None

    def __init__(self, kb_name: str, embed_model: str, vector_name: str, device: str = 'cpu') -> None:
        self.device = device
        self.kb_name = kb_name
        self.embed_model = embed_model
        self.vector_name = vector_name
        self._vs_path = os.path.join(kb_name, "vector_store", vector_name)


    def save_vector_store(self):
        raise NotImplementedError

    def get_doc_by_ids(self, ids: List[str]) -> List[DocumentWithVSId]:

        raise NotImplementedError

    def del_doc_by_ids(self, ids: List[str]) -> bool:

        raise NotImplementedError


    def do_create_kb(self):

        raise NotImplementedError


    def do_search(
            self,
            query: str,
            top_k: int,
            score_threshold: float = 1,
    ) -> List[DocumentWithVSId]:
        raise NotImplementedError

    def do_add_doc(
            self,
            docs: List[DocumentWithVSId],
            **kwargs,
    ) -> List[Dict]:

        raise NotImplementedError

    def do_clear_vs(self):

        raise NotImplementedError



